


cdef struct size_t_vector:
    size_t* v
    size_t used
    size_t size

cdef size_t_vector* make_size_t_vector_with_size(size_t size) nogil
cdef size_t_vector* make_size_t_vector() nogil
cdef int size_t_vector_resize(size_t_vector* vec) nogil
cdef int size_t_vector_append(size_t_vector* vec, size_t value) nogil
cdef int size_t_vector_reserve(size_t_vector* vec, size_t new_size) nogil
cdef void size_t_vector_reset(size_t_vector* vec) nogil

cdef void free_size_t_vector(size_t_vector* vec) nogil
cdef void print_size_t_vector(size_t_vector* vec) nogil

cdef class SizeTVector(object):
    cdef __cythonbufferdefaults__ = {'ndim' : 1, 'mode':'c'}

    cdef:
        size_t_vector* impl
        int flags

    cdef int allocate_storage(self) nogil
    cdef int allocate_storage_with_size(self, size_t size) nogil

    cdef int free_storage(self) nogil
    cdef bint get_should_free(self) nogil
    cdef void set_should_free(self, bint flag) nogil

    cdef size_t* get_data(self) nogil

    @staticmethod
    cdef SizeTVector _create(size_t size)

    @staticmethod
    cdef SizeTVector wrap(size_t_vector* vector)

    cdef size_t get(self, size_t i) nogil
    cdef void set(self, size_t i, size_t value) nogil
    cdef size_t size(self) nogil
    cdef int cappend(self, size_t value) nogil

    cdef SizeTVector _slice(self, object slice_spec)

    cpdef SizeTVector copy(self)

    cpdef int append(self, object value) except *
    cpdef int extend(self, object values) except *

    cpdef int reserve(self, size_t size) nogil

    cpdef int fill(self, size_t value) nogil


    cpdef void qsort(self, bint reverse=?) nogil

    cpdef object _to_python(self, size_t value)
    cpdef size_t _to_c(self, object value) except *


import six
import abc


def simple_repr(self):  # pragma: no cover
    template = "{self.__class__.__name__}({d})"
    d = [
        "%s=%r" % (k, v) if v is not self else "(...)" for k, v in sorted(
            self.__dict__.items(), key=lambda x: x[0])
        if (not k.startswith("_") and not callable(v)) and not (k == "signal")]
    return template.format(self=self, d=', '.join(d))


class Base(object):
    __repr__ = simple_repr


@six.add_metaclass(abc.ABCMeta)
class PeakLike(object):

    @classmethod
    def __subclasshook__(cls, C):
        if cls is PeakLike:
            mz = any(tuple(getattr(B, 'mz', None) is not None for B in C.mro()))
            intensity = any(tuple(getattr(B, 'intensity', None) is not None for B in C.mro()))
            return mz and intensity
        return NotImplemented

    @classmethod
    def is_a(cls, obj):
        val = isinstance(obj, cls)
        if val:
            return val
        val = hasattr(obj, 'mz') and hasattr(obj, 'intensity')
        return val





cdef struct interval_t:
    size_t start
    size_t end


cdef bint interval_overlaps(interval_t* self, interval_t* other) nogil
cdef bint interval_contains(interval_t* self, size_t point) nogil


cdef struct interval_t_vector:
    interval_t* v
    size_t used
    size_t size

cdef int initialize_interval_vector_t(interval_t_vector* vec, size_t size) nogil

cdef interval_t_vector* make_interval_t_vector_with_size(size_t size) nogil
cdef interval_t_vector* make_interval_t_vector() nogil
cdef int interval_t_vector_resize(interval_t_vector* vec) nogil
cdef int interval_t_vector_append(interval_t_vector* vec, interval_t value) nogil
cdef int interval_t_vector_reserve(interval_t_vector* vec, size_t new_size) nogil
cdef void interval_t_vector_reset(interval_t_vector* vec) nogil

cdef void free_interval_t_vector(interval_t_vector* vec) nogil
cdef void print_interval_t_vector(interval_t_vector* vec) nogil

cdef class IntervalVector(object):
    cdef __cythonbufferdefaults__ = {'ndim' : 1, 'mode':'c'}

    cdef:
        interval_t_vector* impl
        int flags

    cdef int allocate_storage(self) nogil
    cdef int allocate_storage_with_size(self, size_t size) nogil

    cdef int free_storage(self) nogil
    cdef bint get_should_free(self) nogil
    cdef void set_should_free(self, bint flag) nogil

    cdef interval_t* get_data(self) nogil

    @staticmethod
    cdef IntervalVector _create(size_t size)

    @staticmethod
    cdef IntervalVector wrap(interval_t_vector* vector)

    cdef interval_t get(self, size_t i) nogil
    cdef void set(self, size_t i, interval_t value) nogil
    cdef size_t size(self) nogil
    cdef int cappend(self, interval_t value) nogil

    cdef IntervalVector _slice(self, object slice_spec)

    cpdef IntervalVector copy(self)

    cpdef int append(self, tuple value) except *
    cpdef int extend(self, object values) except *

    cpdef int reserve(self, size_t size) nogil

    cpdef int fill(self, interval_t value) nogil



    cpdef object _to_python(self, interval_t value)
    cpdef interval_t _to_c(self, object value) except *


# cython: embedsignature=True
#

from libc.math cimport sqrt, exp, pow

from ms_peak_picker._c.double_vector cimport (
    make_double_vector_with_size, make_double_vector,
    double_vector_append, reset_double_vector,
    free_double_vector, DoubleVector)

cimport cython
cimport numpy as np
import numpy as np

np.import_array()


cdef double pi = np.pi


@cython.cdivision(True)
cdef int mean_var(double* x, size_t n, double* mean, double* variance) nogil:
    cdef:
        size_t i
        double acc, m

    acc = 0
    for i in range(n):
        acc += x[i]
    m = mean[0] = acc / n

    acc = 0
    for i in range(n):
        acc += pow(x[i] - m, 2)
    variance[0] = acc / n
    return 0


@cython.cdivision(True)
cdef int vgauss(double* x, size_t n, double mean, double variance, double scale, DoubleVector* out) nogil:
    cdef:
        size_t i
        int err
        double val, a, b

    a = 1. / sqrt(2 * pi * variance)
    b = scale * variance

    for i in range(n):
        err = double_vector_append(out, a * exp((-pow(x[i] - mean, 2)) / b))
        if err != 0:
            return err



@cython.cdivision(True)
@cython.boundscheck(False)
cpdef np.ndarray[double, ndim=1, mode='c'] gaussian_smooth(
        np.ndarray[double, ndim=1, mode='c'] x,
        np.ndarray[double, ndim=1, mode='c'] y, double width=0.05):
    cdef:
        np.ndarray[double, ndim=1, mode='c'] smoothed
        size_t i, j, n, low_edge, high_edge, delta
        double center, spread, xj, thresh, acc, weights_acc
        double* x_part
        double* y_part
        DoubleVector* weights_vector

    smoothed = np.zeros_like(y)

    n = x.shape[0]
    weights_vector = make_double_vector()
    for i in range(n):
        thresh = x[i] - width
        j = i
        while j > 0 and x[j] > thresh:
            j -= 1
        low_edge = j
        j = i
        thresh = x[i] + width
        while j < n and x[j] < thresh:
            j += 1
        high_edge = j
        if high_edge - 1 == low_edge or high_edge == low_edge:
            smoothed[i] = y[i]
            continue
        delta = high_edge - low_edge
        x_part = &x[low_edge]
        y_part = &y[low_edge]
        mean_var(x_part, delta, &center, &spread)
        reset_double_vector(weights_vector)
        vgauss(x_part, delta, center, spread, 2.0, weights_vector)
        acc = 0
        weights_acc = 0
        for j in range(delta):
            acc += y_part[j] * weights_vector.v[j]
            weights_acc += weights_vector.v[j]
        smoothed[i] = acc / weights_acc
    free_double_vector(weights_vector)
    return smoothed



cimport cython
import numpy as np
cimport numpy as np
np.import_array()

from libc cimport math
from libc.math cimport fabs

from ms_peak_picker._c.peak_set cimport FittedPeak

from ms_peak_picker._c.peak_statistics cimport (
    find_signal_to_noise, find_left_width, find_right_width,
    quadratic_fit, lorentzian_fit, peak_area)

from ms_peak_picker._c.search cimport get_nearest_binary, get_nearest

from numpy.linalg import LinAlgError
import logging

logger = logging.getLogger("peak_picker")
info = logger.info
debug = logger.debug


cdef str CENTROID = 'centroid'
cdef str PROFILE = 'profile'


cdef dict fit_type_map = {
    "quadratic": PeakFit.quadratic,
    "gaussian": PeakFit.quadratic,
    "lorenztian": PeakFit.lorentzian,
    "lorentzian": PeakFit.lorentzian,
    "apex": PeakFit.apex,
    PeakFit.apex: PeakFit.apex,
    PeakFit.quadratic: PeakFit.quadratic,
    PeakFit.lorentzian: PeakFit.lorentzian
}


cdef dict peak_mode_map = {
    CENTROID: PeakMode.centroid,
    PROFILE: PeakMode.profile,
    PeakMode.profile: PeakMode.profile,
    PeakMode.centroid: PeakMode.centroid
}



@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline bint isclose(cython.floating x, cython.floating y, cython.floating rtol=1.e-5, cython.floating atol=1.e-8) nogil:
    return math.fabs(x-y) <= (atol + rtol * math.fabs(y))


@cython.nonecheck(False)
@cython.cdivision(True)
cdef inline bint aboutzero(cython.floating x) nogil:
    return isclose(x, 0)


@cython.boundscheck(False)
cpdef bint is_increasing(np.ndarray[cython.floating, ndim=1, mode='c'] mz_array):
    """Test whether the values in `mz_array` are increasing.

    Occaisionally, the m/z array is not completely sorted. This should
    efficiently check if the array is indeed not in the correct order and
    a more expensive sort and re-index operation needs to be performed.

    Parameters
    ----------
    mz_array : :class:`np.ndarray`
        The array to test.

    Returns
    -------
    bool:
        Whether the array is strictly increasing or not.
    """
    cdef:
        size_t i, n
        cython.floating a, b

    n = mz_array.shape[0]
    if n <= 1:
        return True
    with nogil:
        a = mz_array[0]
        for i in range(1, n):
            b = mz_array[i]
            if a >= b:
                return False
            a = b
        return True


@cython.final
@cython.freelist(1000)
cdef class PartialPeakFitState(object):

    """Stores partial state for the peak currently being picked by a :class:`PeakProcessor`
    instance.

    Rather than storing this information directly in the :class:`PeakProcessor` object,
    this separates the state of the current peak from the state of the peak picking process,
    while providing a simple way to clear the current peak's data.

    Attributes
    ----------
    full_width_at_half_max : float
        The complete full width at half max of the current peak
    left_width : float
        The left width at half max of the current peak
    right_width : float
        The right width at half max of the current peak
    set : bool
        Whether or not the current peak has any stored data
    signal_to_noise : float
        The signal to noise ratio of the current peak
    """

    def __init__(self, left_width=-1, right_width=-1, full_width_at_half_max=-1, signal_to_noise=-1):
        self.set = left_width != -1
        self.left_width = left_width
        self.right_width = right_width
        self.full_width_at_half_max = full_width_at_half_max
        self.signal_to_noise = signal_to_noise

    cpdef int reset(self) nogil:
        """Resets all the data in the object to initial configuration
        """
        self.set = False
        self.left_width = -1
        self.right_width = -1
        self.full_width_at_half_max = -1
        self.signal_to_noise = -1
        return 0


cdef class PeakProcessor(object):

    def __init__(self, fit_type='quadratic', peak_mode=PROFILE, signal_to_noise_threshold=1, intensity_threshold=1,
                 threshold_data=False, verbose=False, integrate=True):
        if fit_type not in fit_type_map:
            raise ValueError("Unknown fit_type %r" % (fit_type,))
        else:
            fit_type = fit_type_map[fit_type]
        if peak_mode not in peak_mode_map:
            raise ValueError("Unknown peak_mode %r" % (peak_mode,))
        else:
            peak_mode = peak_mode_map[peak_mode]

        self._signal_to_noise_threshold = 0
        self._intensity_threshold = 0

        self.fit_type = fit_type

        self.background_intensity = 1
        self.threshold_data = threshold_data
        self.signal_to_noise_threshold = signal_to_noise_threshold
        self.intensity_threshold = intensity_threshold

        self.peak_mode = peak_mode
        self.verbose = verbose
        self.integrate = integrate

        self.partial_fit_state = PartialPeakFitState()

        self.peak_data = []

    cpdef double get_signal_to_noise_threshold(self):
        return self._signal_to_noise_threshold

    cpdef object set_signal_to_noise_threshold(self, double signal_to_noise_threshold):
        self._signal_to_noise_threshold = signal_to_noise_threshold

        if self.threshold_data:
            if self.get_signal_to_noise_threshold() != 0:
                self.background_intensity = self.get_intensity_threshold() / self.get_signal_to_noise_threshold()
            else:
                self.background_intensity = 1.

    signal_to_noise_threshold = property(
        get_signal_to_noise_threshold, set_signal_to_noise_threshold)

    cpdef double get_intensity_threshold(self):
        return self._intensity_threshold

    cpdef object set_intensity_threshold(self, double intensity_threshold):
        self._intensity_threshold = intensity_threshold
        if self.threshold_data:
            if self.get_signal_to_noise_threshold() != 0:
                self.background_intensity = intensity_threshold / self.get_signal_to_noise_threshold()
            elif intensity_threshold != 0:
                self.background_intensity = intensity_threshold
            else:
                self.background_intensity = 1.

    intensity_threshold = property(
        get_intensity_threshold, set_intensity_threshold)

    def discover_peaks(self, np.ndarray[cython.floating, ndim=1, mode='c'] mz_array, np.ndarray[cython.floating, ndim=1, mode='c'] intensity_array,
                       start_mz=None, stop_mz=None, start_index=None, stop_index=None):
        """Carries out the peak picking process on `mz_array` and `intensity_array`. All
        peaks picked are appended to :attr:`peak_data`.

        Parameters
        ----------
        mz_array : np.ndarray
            The m/z values to pick peaks from
        intensity_array : np.ndarray
            The intensity values to pick peaks from
        start_mz : float, optional
            The minimum m/z to pick peaks above
        stop_mz : float, optional
            The maximum m/z to pick peaks below

        Returns
        -------
        int
            The current number of peaks accumulated
        """
        cdef:
            double _start_mz, _stop_mz
            Py_ssize_t _start_index, _stop_index

        if start_mz is None:
            _start_mz = -1
        else:
            _start_mz = start_mz
        if stop_mz is None:
            _stop_mz = -1
        else:
            _stop_mz = stop_mz
        if start_index is None:
            _start_index = 0
        else:
            _start_index = start_index
        if stop_index is None:
            _stop_index = 0
        else:
            _stop_index = stop_index
        return self._discover_peaks(mz_array, intensity_array, _start_mz, _stop_mz, _start_index, _stop_index)

    @cython.nonecheck(False)
    @cython.cdivision(True)
    @cython.boundscheck(False)
    cpdef size_t _discover_peaks(self, np.ndarray[cython.floating, ndim=1, mode='c'] mz_array, np.ndarray[cython.floating, ndim=1, mode='c'] intensity_array,
                                 double start_mz, double stop_mz, Py_ssize_t start_index=0, Py_ssize_t stop_index=0):
        """Carries out the peak picking process on `mz_array` and `intensity_array`. All
        peaks picked are appended to :attr:`peak_data`.

        Parameters
        ----------
        mz_array : np.ndarray
            The m/z values to pick peaks from
        intensity_array : np.ndarray
            The intensity values to pick peaks from
        start_mz : float, optional
            The minimum m/z to pick peaks above
        stop_mz : float, optional
            The maximum m/z to pick peaks below

        Returns
        -------
        int
            The current number of peaks accumulated
        """
        cdef:
            Py_ssize_t size, ihigh, ilow, index
            list peak_data
            bint verbose, is_centroid
            double intensity_threshold, signal_to_noise_threshold, signal_to_noise
            double current_intensity, last_intensity, next_intensity, sum_intensity
            double low_intensity, high_intensity
            double full_width_at_half_max, current_mz, mz
            size_t offset

            FittedPeak peak

        size = len(intensity_array) - 1

        if size < 1:
            return 0

        if start_mz <= -1:
            start_mz = mz_array[0]
        if stop_mz <= -1:
            stop_mz = mz_array[len(mz_array) - 1]

        peak_data = []

        verbose = self.verbose
        is_centroid = self.peak_mode == PeakMode.centroid

        intensity_threshold = self.intensity_threshold
        signal_to_noise_threshold = self.signal_to_noise_threshold

        if start_index == stop_index == 0:
            start_index = get_nearest_binary(mz_array, start_mz, 0, size)
            stop_index = get_nearest_binary(mz_array, stop_mz, start_index, size)

        if start_index <= 0 and not is_centroid:
            start_index = 1
        elif start_index < 0 and is_centroid:
            start_index = 0
        if stop_index >= size - 1 and not is_centroid:
            stop_index = size - 1
        elif stop_index >= size and is_centroid:
            stop_index = size

        index = start_index
        while index <= stop_index:
        # for index in range(start_index, stop_index + 1):
            self.partial_fit_state.reset()
            full_width_at_half_max = -1
            current_intensity = intensity_array[index]

            current_mz = mz_array[index]

            # If we are dealing with pre-centroided data, just walk down the array making each
            # point a FittedPeak.
            if is_centroid:
                if current_intensity <= 0:
                    index += 1
                    continue
                mz = mz_array[index]
                signal_to_noise = current_intensity / (intensity_threshold or 1.0)
                full_width_at_half_max = 0.025

                if signal_to_noise > signal_to_noise_threshold:
                    peak = FittedPeak._create(
                        mz, current_intensity, signal_to_noise,
                        full_width_at_half_max, full_width_at_half_max / 2.,
                        full_width_at_half_max / 2., len(peak_data), index, area)
                    peak_data.append(peak)
            # Otherwise, carry out the peak finding and fitting procedure.
            else:
                last_intensity = intensity_array[index - 1]
                next_intensity = intensity_array[index + 1]

                # Three point peak picking. Check if the peak is greater than
                # both the previous and next points
                if (current_intensity >= last_intensity and current_intensity >= next_intensity and
                        current_intensity >= intensity_threshold):
                    signal_to_noise = 0.
                    if not self.threshold_data:
                        signal_to_noise = find_signal_to_noise(
                            current_intensity, intensity_array, index)
                    else:
                        signal_to_noise = current_intensity / \
                            self.background_intensity

                    # Run Full-Width Half-Max algorithm to try to improve SNR
                    if signal_to_noise < signal_to_noise_threshold:
                        try:
                            full_width_at_half_max = self.find_full_width_at_half_max(
                                index,
                                mz_array, intensity_array, signal_to_noise)
                        except LinAlgError:
                            full_width_at_half_max = 0
                        if 0 < full_width_at_half_max < 0.5:
                            ilow = get_nearest_binary(
                                mz_array, current_mz - self.partial_fit_state.left_width, 0, index)
                            ihigh = get_nearest_binary(
                                mz_array, current_mz + self.partial_fit_state.right_width, index, stop_index)

                            low_intensity = intensity_array[ilow]
                            high_intensity = intensity_array[ihigh]

                            sum_intensity = low_intensity + high_intensity

                            if sum_intensity:
                                signal_to_noise = (
                                    2. * current_intensity) / sum_intensity
                            else:
                                signal_to_noise = 10.

                    self.partial_fit_state.signal_to_noise = signal_to_noise
                    # Found a putative peak, fit it
                    if signal_to_noise >= signal_to_noise_threshold:
                        fitted_mz = self.fit_peak(
                            index, mz_array, intensity_array)
                        if verbose:
                            debug(
                                "Considering peak at %d with fitted m/z %r", index, fitted_mz)
                        if full_width_at_half_max == -1:
                            try:
                                full_width_at_half_max = self.find_full_width_at_half_max(
                                    index,
                                    mz_array, intensity_array, signal_to_noise)
                            except LinAlgError:
                                full_width_at_half_max = 0

                        if full_width_at_half_max > 0:
                            if self.integrate:
                                area = self.area(
                                    mz_array, intensity_array, fitted_mz, full_width_at_half_max, index)
                            else:
                                area = current_intensity
                            if full_width_at_half_max > 1.:
                                full_width_at_half_max = 1.

                            if signal_to_noise > current_intensity:
                                signal_to_noise = current_intensity

                            peak = FittedPeak._create(
                                fitted_mz, current_intensity, signal_to_noise,
                                full_width_at_half_max, self.partial_fit_state.left_width,
                                self.partial_fit_state.right_width, len(peak_data), index, area)
                            peak_data.append(peak)

                        # Move past adjacent equal-height peaks. This seems to fail when the spacing
                        # between m/z coordinates is very, very small (1e-3)
                        incremented = False
                        while index < size and isclose(intensity_array[index + 1], current_intensity):
                            incremented = True
                            if verbose:
                                debug("Advancing over equal intensity point %d @ %f", index, current_intensity)
                            index += 1
                        # if index > 0 and index < size and incremented:
                        #     index -= 1
            index += 1
        self.peak_data.extend(peak_data)
        return len(peak_data)

    cpdef double find_full_width_at_half_max(self, Py_ssize_t index, np.ndarray[cython.floating, ndim=1, mode='c'] mz_array,
                                             np.ndarray[cython.floating, ndim=1, mode='c'] intensity_array, double signal_to_noise):
        """Calculate full-width-at-half-max for a peak centered at `index` from
        `mz_array` and `intensity_array`, using the `signal_to_noise` to detect
        when to stop searching.

        This method will set attributes on :attr:`partial_fit_state`.

        Parameters
        ----------
        index : int
            The index of peak apex
        mz_array : np.ndarray
            The m/z array to search in
        intensity_array : np.ndarray
            The intensity array to search in
        signal_to_noise : float
            The signal-to-noise ratio for this peak

        Returns
        -------
        float
            The symmetric full-width-at-half-max
        """
        cdef:
            double left, right, fwhm
        try:
            left = find_left_width(
                mz_array, intensity_array, index, signal_to_noise)
        except np.linalg.LinAlgError:
            left = 1e-7
        try:
            right = find_right_width(
                mz_array, intensity_array, index, signal_to_noise)
        except np.linalg.LinAlgError:
            right = 1e-7

        if left < 1e-6:
            left = right
        elif right < 1e-6:
            right = left
        if right < 1e-6 and left < 1e-6:
            left = right = 0.15
        fwhm = left + right
        self.partial_fit_state.left_width = left
        self.partial_fit_state.right_width = right
        self.partial_fit_state.full_width_at_half_max = fwhm

        return fwhm

    cpdef double fit_peak(self, Py_ssize_t index,
                          np.ndarray[cython.floating, ndim=1, mode='c'] mz_array,
                          np.ndarray[cython.floating, ndim=1, mode='c'] intensity_array):
        """Performs the peak shape fitting procedure.

        Parameters
        ----------
        index : int
            The index to start the peak fit from
        mz_array : np.ndarray
            The m/z array to search in
        intensity_array : np.ndarray
            The intensity array to search in

        Returns
        -------
        float
            m/z of the fitted peak center
        """
        if self.fit_type == PeakFit.apex:
            return mz_array[index]
        elif self.fit_type == PeakFit.quadratic:
            return quadratic_fit(mz_array, intensity_array, index)
        elif self.fit_type == PeakFit.lorentzian:
            full_width_at_half_max = self.find_full_width_at_half_max(
                index, mz_array, intensity_array,
                self.partial_fit_state.signal_to_noise)
            if full_width_at_half_max != 0:
                return lorentzian_fit(mz_array, intensity_array, index, full_width_at_half_max)
            return mz_array[index]

        return 0.0

    cpdef double area(self,
                      np.ndarray[cython.floating, ndim=1, mode='c'] mz_array,
                      np.ndarray[cython.floating, ndim=1, mode='c'] intensity_array,
                      double mz, double full_width_at_half_max, Py_ssize_t index):
        """Integrate the peak found at `index` with width `full_width_at_half_max`,
        centered at `mz`.

        Parameters
        ----------
        mz_array : np.ndarray
            The m/z array to search in
        intensity_array : np.ndarray
            The intensity array to search in
        mz : float
            The center m/z to start from
        full_width_at_half_max : float
            The width to use when extracting
            the range of points to integrate
        index : int
            The index to start the search from

        Returns
        -------
        float
            The integrated peak area
        """
        lo = get_nearest(mz_array, mz - full_width_at_half_max, index)
        hi = get_nearest(mz_array, mz + full_width_at_half_max, index)
        return peak_area(mz_array, intensity_array, lo, hi)

    def __iter__(self):
        for peak in self.peak_data:
            yield peak




cdef class PeakBase(object):
    cdef:
        public double mz
        public double intensity
        public double area

    cpdef PeakBase clone(self)
    cpdef PeakBase copy(self)


cdef class FittedPeak(PeakBase):
    cdef:
        public double signal_to_noise
        public double full_width_at_half_max
        public double left_width
        public double right_width
        public long peak_count
        public long index

    cpdef bint _eq(self, FittedPeak other)

    @staticmethod
    cdef FittedPeak _create(double mz, double intensity, double signal_to_noise,
                            double full_width_at_half_max, double left_width,
                            double right_width, long peak_count, long index,
                            double area)

cdef class PeakSet(object):
    cdef:
        public tuple peaks
        public bint indexed

    @staticmethod
    cdef PeakSet _create(tuple peaks)

    cpdef reindex(self)
    cpdef size_t _index(self)

    cdef FittedPeak _get_nearest_peak(self, double mz, double* errout)

    cdef FittedPeak _has_peak(self, double mz, double tolerance=*)

    cpdef PeakSet clone(self)
    cpdef PeakSet copy(self)

    cdef PeakSet _between(self, double m1, double m2)
    cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp)

    cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)

    cdef FittedPeak getitem(self, size_t i)

    cdef size_t get_size(self)

    cpdef bint _eq(self, PeakSet other)


cpdef FittedPeak binary_search_ppm_error(tuple array, double value, double tolerance)
cdef FittedPeak _binary_search_ppm_error(tuple array, double value, size_t lo, size_t hi, double tolerance)
cdef FittedPeak _binary_search_nearest_match(tuple array, double value, size_t lo, size_t hi, double* errout)


cdef struct index_cell:
    double center_value
    size_t start
    size_t end


cdef struct index_list:
    index_cell* index
    size_t size
    double low
    double high


cdef int check_index(index_list* index) nogil


cdef size_t INTERVAL_INDEX_SIZE


cdef class PeakSetIndexed(PeakSet):
    cdef:
        double* mz_index
        index_list* interval_index

    cpdef _allocate_index(self, size_t interval_index_size)

    cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=*, size_t hint_start=*, size_t hint_end=*)

cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n)
cdef size_t double_binary_search_nearest_match(double* array, double value, size_t n)
cdef size_t double_binary_search_ppm_with_hint(double* array, double value, double tolerance, size_t n, size_t hint)



'''
A Peak Picker/Fitter adapted from Decon2LS's DeconEngine
'''

import numpy as np
from numpy.linalg import LinAlgError

from .peak_statistics import (
    find_signal_to_noise, quadratic_fit, lorentzian_fit,
    peak_area, find_left_width, find_right_width)

from .search import get_nearest_binary, get_nearest
from .peak_set import FittedPeak, PeakSet
from .peak_index import PeakIndex
from .scan_filter import transform
from .utils import Base


import logging

logger = logging.getLogger("peak_picker")
info = logger.info
debug = logger.debug


fit_type_map = {
    "quadratic": "quadratic",
    "gaussian": "quadratic",
    "lorenztian": "lorentzian",
    "lorentzian": "lorentzian",
    "apex": "apex"
}


CENTROID = "centroid"
PROFILE = "profile"


peak_mode_map = {
    CENTROID: CENTROID,
    PROFILE: PROFILE
}


class PartialPeakFitState(Base):
    """Stores partial state for the peak currently being picked by a :class:`PeakProcessor`
    instance.

    Rather than storing this information directly in the :class:`PeakProcessor` object,
    this separates the state of the current peak from the state of the peak picking process,
    while providing a simple way to clear the current peak's data.

    Attributes
    ----------
    full_width_at_half_max : float
        The complete full width at half max of the current peak
    left_width : float
        The left width at half max of the current peak
    right_width : float
        The right width at half max of the current peak
    set : bool
        Whether or not the current peak has any stored data
    signal_to_noise : float
        The signal to noise ratio of the current peak
    """

    def __init__(self, left_width=-1, right_width=-1, full_width_at_half_max=-1, signal_to_noise=-1):
        self.set = left_width != -1
        self.left_width = left_width
        self.right_width = right_width
        self.full_width_at_half_max = full_width_at_half_max
        self.signal_to_noise = signal_to_noise

    def reset(self):
        """Resets all the data in the object to initial configuration
        """
        self.set = False
        self.left_width = -1
        self.right_width = -1
        self.full_width_at_half_max = -1
        self.signal_to_noise = -1


class PeakProcessor(object):
    """Directs the peak picking process, encapsulating the apex finding,
    peak fitting, and signal-to-noise estimation tasks.

    Attributes
    ----------
    background_intensity : float
        A static background intensity to use when estimating signal-to-noise
        ratio.
    fit_type : str
        The type of peak to fit
    intensity_threshold : float
        The minimum intensity required to accept a peak
    partial_fit_state : PartialPeakFitState
        A stateful container of measurements for the peak currently
        being fitted
    peak_data : list
        A list of :class:`ms_peak_picker.peak_set.FittedPeak` instances
    peak_mode : str
        Whether the peaks being picked are in profile mode or already centroided
        and just need to be passed directly into :class:`FittedPeak` instances
    signal_to_noise_threshold : float
        The minimum signal-to-noise ratio required to accept a peak fit
    threshold_data : bool
        Whether or not to enforce a signal-to-noise and intensity threhsold
    verbose : bool
        Whether to log additional diagnostic information
    """

    def __init__(self, fit_type='quadratic', peak_mode=PROFILE, signal_to_noise_threshold=1, intensity_threshold=1,
                 threshold_data=False, verbose=False, integrate=True):
        if fit_type not in fit_type_map:
            raise ValueError("Unknown fit_type %r" % (fit_type,))
        if peak_mode not in peak_mode_map:
            raise ValueError("Unknown peak_mode %r" % (peak_mode,))
        # normalize fit_type
        fit_type = fit_type_map[fit_type]
        self._signal_to_noise_threshold = 0
        self._intensity_threshold = 0

        self.fit_type = fit_type

        self.background_intensity = 1
        self.threshold_data = threshold_data
        self.signal_to_noise_threshold = signal_to_noise_threshold
        self.intensity_threshold = intensity_threshold

        self.peak_mode = peak_mode
        self.verbose = verbose
        self.integrate = integrate

        self.partial_fit_state = PartialPeakFitState()

        self.peak_data = []

    def get_signal_to_noise_threshold(self):
        return self._signal_to_noise_threshold

    def set_signal_to_noise_threshold(self, signal_to_noise_threshold):
        self._signal_to_noise_threshold = signal_to_noise_threshold

        if self.threshold_data:
            if self.signal_to_noise_threshold != 0:
                self.background_intensity = (
                    self.intensity_threshold / float(self.signal_to_noise_threshold))
            else:
                self.background_intensity = 1.

    signal_to_noise_threshold = property(
        get_signal_to_noise_threshold, set_signal_to_noise_threshold)

    def get_intensity_threshold(self):
        return self._intensity_threshold

    def set_intensity_threshold(self, intensity_threshold):
        self._intensity_threshold = intensity_threshold
        if self.threshold_data:
            if self.signal_to_noise_threshold != 0:
                self.background_intensity = intensity_threshold / \
                    float(self.signal_to_noise_threshold)
            elif intensity_threshold != 0:
                self.background_intensity = intensity_threshold
            else:
                self.background_intensity = 1.

    intensity_threshold = property(
        get_intensity_threshold, set_intensity_threshold)

    def discover_peaks(self, mz_array, intensity_array, start_mz=None, stop_mz=None):
        """Carries out the peak picking process on `mz_array` and `intensity_array`. All
        peaks picked are appended to :attr:`peak_data`.

        Parameters
        ----------
        mz_array : np.ndarray
            The m/z values to pick peaks from
        intensity_array : np.ndarray
            The intensity values to pick peaks from
        start_mz : float, optional
            The minimum m/z to pick peaks above
        stop_mz : float, optional
            The maximum m/z to pick peaks below

        Returns
        -------
        int
            The current number of peaks accumulated
        """
        size = len(intensity_array) - 1

        if size < 1:
            return 0

        if start_mz is None:
            start_mz = mz_array[0]
        if stop_mz is None:
            stop_mz = mz_array[len(mz_array) - 1]

        peak_data = []

        verbose = self.verbose

        intensity_threshold = self.intensity_threshold
        signal_to_noise_threshold = self.signal_to_noise_threshold

        start_index = get_nearest_binary(mz_array, start_mz, 0, size)
        stop_index = get_nearest_binary(mz_array, stop_mz, start_index, size)

        if start_index <= 0 and self.peak_mode != CENTROID:
            start_index = 1
        elif start_index < 0 and self.peak_mode == CENTROID:
            start_index = 0
        if stop_index >= size - 1 and self.peak_mode != CENTROID:
            stop_index = size - 1
        elif stop_index >= size and self.peak_mode == CENTROID:
            stop_index = size

        for index in range(start_index, stop_index + 1):
            self.partial_fit_state.reset()
            full_width_at_half_max = -1
            current_intensity = intensity_array[index]

            current_mz = mz_array[index]

            if self.peak_mode == CENTROID:
                if current_intensity <= 0:
                    continue
                mz = mz_array[index]
                signal_to_noise = current_intensity / intensity_threshold
                full_width_at_half_max = 0.025
                if signal_to_noise > signal_to_noise_threshold:
                    peak_data.append(FittedPeak(mz, current_intensity, signal_to_noise, len(
                        peak_data), index, full_width_at_half_max, current_intensity))
            else:
                last_intensity = intensity_array[index - 1]
                next_intensity = intensity_array[index + 1]

                # Three point peak picking. Check if the peak is greater than
                # both the previous and next points
                if (current_intensity >= last_intensity and current_intensity >= next_intensity and
                        current_intensity >= intensity_threshold):
                    signal_to_noise = 0.
                    if not self.threshold_data:
                        signal_to_noise = find_signal_to_noise(
                            current_intensity, intensity_array, index)
                    else:
                        signal_to_noise = current_intensity / \
                            float(self.background_intensity)

                    # Run Full-Width Half-Max algorithm to try to improve SNR
                    if signal_to_noise < signal_to_noise_threshold:
                        try:
                            full_width_at_half_max = self.find_full_width_at_half_max(
                                index,
                                mz_array, intensity_array, signal_to_noise)
                        except LinAlgError:
                            full_width_at_half_max = 0
                        if 0 < full_width_at_half_max < 0.5:
                            ilow = get_nearest_binary(
                                mz_array, current_mz - self.partial_fit_state.left_width, 0, index)
                            ihigh = get_nearest_binary(
                                mz_array, current_mz + self.partial_fit_state.right_width, index, stop_index)

                            low_intensity = intensity_array[ilow]
                            high_intensity = intensity_array[ihigh]

                            sum_intensity = low_intensity + high_intensity

                            if sum_intensity:
                                signal_to_noise = (
                                    2. * current_intensity) / sum_intensity
                            else:
                                signal_to_noise = 10.

                    self.partial_fit_state.signal_to_noise = signal_to_noise
                    # Found a putative peak, fit it
                    if signal_to_noise >= signal_to_noise_threshold:
                        fitted_mz = self.fit_peak(
                            index, mz_array, intensity_array)
                        if verbose:
                            debug(
                                "Considering peak at %d with fitted m/z %r", index, fitted_mz)
                        if full_width_at_half_max == -1:
                            try:
                                full_width_at_half_max = self.find_full_width_at_half_max(
                                    index,
                                    mz_array, intensity_array, signal_to_noise)
                            except LinAlgError:
                                full_width_at_half_max = 0

                        if full_width_at_half_max > 0:
                            if self.integrate:
                                area = self.area(
                                    mz_array, intensity_array, fitted_mz, full_width_at_half_max, index)
                            else:
                                area = current_intensity
                            if full_width_at_half_max > 1.:
                                full_width_at_half_max = 1.

                            if signal_to_noise > current_intensity:
                                signal_to_noise = current_intensity

                            peak_data.append(FittedPeak(
                                fitted_mz, current_intensity, signal_to_noise,
                                len(peak_data), index, full_width_at_half_max, area,
                                self.partial_fit_state.left_width, self.partial_fit_state.right_width))

                        # Move past adjacent equal-height peaks
                        incremented = False
                        while index < size and intensity_array[index] == current_intensity:
                            incremented = True
                            index += 1
                        if index > 0 and index < size and incremented:
                            index -= 1
        self.peak_data.extend(peak_data)
        return len(peak_data)

    def find_full_width_at_half_max(self, index, mz_array, intensity_array, signal_to_noise):
        """Calculate full-width-at-half-max for a peak centered at `index` from
        `mz_array` and `intensity_array`, using the `signal_to_noise` to detect
        when to stop searching.

        This method will set attributes on :attr:`partial_fit_state`.

        Parameters
        ----------
        index : int
            The index of peak apex
        mz_array : np.ndarray
            The m/z array to search in
        intensity_array : np.ndarray
            The intensity array to search in
        signal_to_noise : float
            The signal-to-noise ratio for this peak

        Returns
        -------
        float
            The symmetric full-width-at-half-max
        """
        try:
            left = find_left_width(
                mz_array, intensity_array, index, signal_to_noise)
        except np.linalg.LinAlgError:
            left = 1e-7
        try:
            right = find_right_width(
                mz_array, intensity_array, index, signal_to_noise)
        except np.linalg.LinAlgError:
            right = 1e-7

        if left < 1e-6:
            left = right
        elif right < 1e-6:
            right = left
        if right < 1e-6 and left < 1e-6:
            left = right = 0.15
        fwhm = left + right
        self.partial_fit_state.left_width = left
        self.partial_fit_state.right_width = right
        self.partial_fit_state.full_width_at_half_max = fwhm

        return fwhm

    def fit_peak(self, index, mz_array, intensity_array):
        """Performs the peak shape fitting procedure.

        Parameters
        ----------
        index : int
            The index to start the peak fit from
        mz_array : np.ndarray
            The m/z array to search in
        intensity_array : np.ndarray
            The intensity array to search in

        Returns
        -------
        float
            m/z of the fitted peak center
        """
        if self.fit_type == "apex":
            return mz_array[index]
        elif self.fit_type == "quadratic":
            return quadratic_fit(mz_array, intensity_array, index)
        elif self.fit_type == "lorentzian":
            full_width_at_half_max = self.find_full_width_at_half_max(
                index, mz_array, intensity_array,
                self.partial_fit_state.signal_to_noise)
            if full_width_at_half_max != 0:
                return lorentzian_fit(mz_array, intensity_array, index, full_width_at_half_max)
            return mz_array[index]

        return 0.0

    def area(self, mz_array, intensity_array, mz, full_width_at_half_max, index):
        """Integrate the peak found at `index` with width `full_width_at_half_max`,
        centered at `mz`.

        Parameters
        ----------
        mz_array : np.ndarray
            The m/z array to search in
        intensity_array : np.ndarray
            The intensity array to search in
        mz : float
            The center m/z to start from
        full_width_at_half_max : float
            The width to use when extracting
            the range of points to integrate
        index : int
            The index to start the search from

        Returns
        -------
        float
            The integrated peak area
        """
        lo = get_nearest(mz_array, mz - full_width_at_half_max, index)
        hi = get_nearest(mz_array, mz + full_width_at_half_max, index)
        return peak_area(mz_array, intensity_array, lo, hi)

    def __iter__(self):
        for peak in self.peak_data:
            yield peak


def pick_peaks(mz_array, intensity_array, fit_type='quadratic', peak_mode=PROFILE,
               signal_to_noise_threshold=1., intensity_threshold=1., threshold_data=False,
               target_envelopes=None, transforms=None, verbose=False,
               start_mz=None, stop_mz=None, integrate=True, target_indices=None):
    """Picks peaks for the given m/z, intensity array pair, producing a centroid-containing
    PeakIndex instance.

    Applies each :class:`.FilterBase` in `transforms` in order to
    `mz_array` and `intensity_array`.

    Creates an instance of :class:`.PeakProcessor` and configures it according to the parameters
    passed. If `target_envelopes` is set, each region is handled by :meth:`.PeakProcessor.discover_peaks`
    otherwise, :meth:`.PeakProcessor.discover_peaks` is invoked with `start_mz` and `stop_mz`.

    Produces a :class:`.PeakIndex`, a fast searchable collection of :class:`.FittedPeak` objects.

    Parameters
    ----------
    mz_array : np.ndarray
        An array of m/z measurements. Will by converted into np.float64
        values
    intensity_array : np.ndarray
        An array of intensity measurements. Will by converted into np.float64
        values
    fit_type : str, optional
        The name of the peak model to use. One of "quadratic", "gaussian", "lorentzian", or "apex"
    peak_mode : str, optional
        Whether peaks are in "profile" mode or are pre"centroid"ed
    signal_to_noise_threshold : float, optional
        Minimum signal-to-noise measurement to accept a peak
    intensity_threshold : float, optional
        Minimum intensity measurement to accept a peak
    threshold_data : bool, optional
        Whether to apply thresholds to the data
    target_envelopes : list, optional
        A sequence of (start m/z, end m/z) pairs, limiting peak picking to only those intervals
    transforms : list, optional
        A list of :class:`scan_filter.FilterBase` instances or callable that
        accepts (mz_array, intensity_array) and returns (mz_array, intensity_array) or
        `str` matching one of the premade names in `scan_filter.filter_register`
    verbose : bool, optional
        Whether to log extra information while picking peaks
    start_mz : float, optional
        A minimum m/z value to start picking peaks from
    stop_mz : None, optional
        A maximum m/z value to stop picking peaks after
    integrate: bool, optional
        Whether to integrate along each peak to calculate the area. Defaults
        to :const:`True`, but the area value for each peak is not usually used
        by downstream algorithms for consistency, so this expensive operation
        can be omitted.

    Returns
    -------
    :class:`PeakIndex`
        Contains all fitted peaks, as well as the transformed m/z and
        intensity arrays
    """
    if transforms is None:
        transforms = []

    mz_array = np.asanyarray(mz_array, dtype=np.float64)
    intensity_array = np.asanyarray(intensity_array, dtype=np.float64)

    if len(mz_array) != len(intensity_array):
        raise ValueError("The m/z array and intensity array must be the same size!")

    # make sure the m/z array is properly sorted
    if not is_increasing(mz_array):
        indexing = np.argsort(mz_array)
        mz_array = mz_array[indexing]
        intensity_array = intensity_array[indexing]

    mz_array, intensity_array = transform(
        mz_array, intensity_array, transforms)

    if len(mz_array) < 1:
        return PeakIndex(mz_array, intensity_array, PeakSet([]))

    processor = PeakProcessor(
        fit_type, peak_mode, signal_to_noise_threshold, intensity_threshold, threshold_data,
        verbose=verbose, integrate=integrate)
    if target_envelopes:
        for start, stop in sorted(target_envelopes):
            processor.discover_peaks(
                mz_array, intensity_array, start_mz=start, stop_mz=stop)
    elif target_indices:
        for start, stop in sorted(target_indices):
            processor.discover_peaks(mz_array, intensity_array, start_index=start, stop_index=stop)
    else:
        processor.discover_peaks(
            mz_array, intensity_array,
            start_mz=start_mz, stop_mz=stop_mz)

    peaks = PeakSet(processor)
    peaks._index()
    return PeakIndex(mz_array, intensity_array, peaks)


def is_increasing(mz_array):
    """Test whether the values in `mz_array` are increasing.

    Occaisionally, the m/z array is not completely sorted. This should
    efficiently check if the array is indeed not in the correct order and
    a more expensive sort and re-index operation needs to be performed.

    Parameters
    ----------
    mz_array : :class:`np.ndarray`
        The array to test.

    Returns
    -------
    bool:
        Whether the array is strictly increasing or not.
    """
    return np.all(mz_array[1:] > mz_array[:-1])

try:
    _has_c = True
    _PartialPeakFitState = PartialPeakFitState
    _PeakProcessor = PeakProcessor
    from ms_peak_picker._c.peak_picker import PartialPeakFitState, PeakProcessor, is_increasing
except ImportError as err:
    print(err)
    _has_c = False



'''
An implementation of the noise filtering methods from MasSpike
'''

from typing import Tuple
import numpy as np


def binsearch(array, x, hint=None):
    n = len(array)
    lo = 0
    hi = n

    while hi != lo:
        mid = (hi + lo) // 2
        y = array[mid]
        err = y - x
        if hi - lo == 1:
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return 0


def between_search(array, lo, hi):
    n = len(array)
    lo_i = binsearch(array, lo)
    hi_i = binsearch(array, hi)

    if lo - array[lo_i] > 0.1:
        if lo_i < n - 1:
            lo_i += 1
    if ((array[hi_i] - hi) > 0.1) and (hi_i - 1) > lo_i:
        if hi_i > 0:
            hi_i -= 1
    if lo_i > hi_i:
        hi_i = lo_i
    return lo_i, hi_i


class Window(object):

    def __init__(self, mz_array, intensity_array, start_index=None, end_index=None,
                 center_mz=None):
        self.mz_array = mz_array
        self.intensity_array = intensity_array
        self.start_index = start_index
        self.end_index = end_index
        self.center_mz = center_mz

    def deduct_intensity(self, value):
        self.intensity_array -= value
        self.intensity_array.clip(min=0, out=self.intensity_array)

    @property
    def mean_intensity(self):
        return self.intensity_array.mean()

    def truncated_mean(self, threshold=0.95):
        try:
            hist_count, hist_values = np.histogram(self.intensity_array)
        except ValueError:
            return 1e-6
        mask = (hist_count.max() * (1 - threshold)) < hist_count
        mean = (hist_values[1:][mask] * hist_count[mask]
                ).sum() / hist_count[mask].sum()
        return mean

    def __repr__(self):
        return "Window(start_index=%d, end_index=%d, mean_intensity=%f, center_mz=%f)" % (
            self.start_index, self.end_index, self.mean_intensity, self.center_mz)


def windowed_spectrum(mz_array, intensity_array, window_size=1.):
    mz_min = mz_array.min()
    mz_max = mz_array.max()

    step_size = window_size / 2.
    center_mz = mz_min + step_size
    center_i = binsearch(mz_array, center_mz, 1)

    windows = []

    niter = 0

    while center_mz < mz_max:
        lo_mz = center_mz - step_size
        hi_mz = center_mz + step_size

        # lo_i = binsearch(mz_array, lo_mz, center_i)
        # hi_i = binsearch(mz_array, hi_mz, center_i)
        lo_i, hi_i = between_search(mz_array, lo_mz, hi_mz)
        mz_window = mz_array[lo_i:hi_i + 1]
        if lo_mz < mz_window.mean() < hi_mz:
            win = Window(mz_window,
                         intensity_array[lo_i:hi_i + 1], lo_i, hi_i)
        else:
            win = Window(np.array([center_mz]), np.array([0.0]), lo_i, hi_i)
        win.center_mz = center_mz
        windows.append(win)

        center_mz = center_mz + window_size
        center_i = binsearch(mz_array, center_mz, center_i)

        niter += 1

    return windows


class NoiseRegion(object):

    def __init__(self, windows, width=10):
        self.windows = windows
        self.width = width
        self.start_index = windows[0].start_index
        self.end_index = windows[-1].end_index

    def __repr__(self):
        return "NoiseRegion(start_index=%d, end_index=%d)" % (
            self.start_index, self.end_index)

    def noise_window(self):
        return min(self.windows, key=lambda x: x.mean_intensity)

    def arrays(self):
        mz_array = np.concatenate([w.mz_array for w in self.windows])
        intensity_array = np.concatenate(
            [w.intensity_array for w in self.windows])
        return mz_array, intensity_array

    def denoise(self, scale=5, maxiter=10):
        if scale == 0:
            return 0
        noise_mean = self.noise_window().truncated_mean() * scale
        for window in self.windows:
            window.deduct_intensity(noise_mean)
        last_mean = noise_mean
        noise_mean = self.noise_window().truncated_mean()
        niter = 1
        while abs(last_mean - noise_mean) > 1e-3 and niter < maxiter:
            niter += 1
            noise_mean = self.noise_window().truncated_mean() * scale
            for window in self.windows:
                window.deduct_intensity(noise_mean)
            last_mean = noise_mean

        return last_mean - noise_mean


def group_windows(windows, width=10):
    step = int(width / 2)
    regions = []
    i = step
    while i < len(windows):
        lo = i - step
        hi = i + step
        reg = NoiseRegion(windows[lo:hi])
        regions.append(reg)
        i += step * 2
    return regions


class FTICRScan(object):

    def __init__(self, mz_array, intensity_array):
        self.mz_array = mz_array.copy()
        self.intensity_array = intensity_array.copy()

        self.windows = None
        self.regions = None

    def __iter__(self):
        yield self.mz_array
        yield self.intensity_array

    def build_windows(self, window_size=1.):
        self.windows = windowed_spectrum(
            self.mz_array, self.intensity_array, window_size=window_size)

    def build_regions(self, region_width=10):
        self.regions = group_windows(self.windows, region_width)

    def denoise(self, window_size=1., region_width=10, scale=5):
        self.build_windows(window_size)
        self.build_regions(region_width)
        for region in self.regions:
            region.denoise(scale)
        return self


try:
    _FTICRScan = FTICRScan
    has_c = True
    from ms_peak_picker._c.fticr_denoising import FTICRScan
except ImportError:
    has_c = False


def denoise(mz_array: np.ndarray, intensity_array: np.ndarray, window_size: float=1., region_width: float=10, scale: float=5) -> Tuple[np.ndarray, np.ndarray]:
    scan = FTICRScan(mz_array, intensity_array)
    denoised = scan.denoise(window_size, region_width, scale)
    x, y = list(denoised)
    assert x.shape == mz_array.shape and y.shape == intensity_array.shape
    return x, y



# cython: embedsignature=True

cimport cython
from cython cimport parallel
cimport numpy as cnp
from libc cimport math
from libc.stdlib cimport malloc, calloc, free
from multiprocessing import cpu_count
import numpy as np

cdef extern from * nogil:
    int printf (const char *template, ...)

from cpython cimport PyFloat_AsDouble, PyInt_AsLong, PyErr_SetString, PyErr_Format
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM
from cpython.tuple cimport PyTuple_GET_SIZE, PyTuple_GET_ITEM
from cpython.slice cimport PySlice_GetIndicesEx

from ms_peak_picker._c.double_vector cimport (
    make_double_vector_with_size, make_double_vector,
    double_vector_append, reset_double_vector,
    free_double_vector, DoubleVector)


from ms_peak_picker._c.size_t_vector cimport (
    make_size_t_vector_with_size, make_size_t_vector,
    size_t_vector_append, free_size_t_vector,
    size_t_vector, size_t_vector_reset, SizeTVector
)


from ms_peak_picker._c.interval_t_vector cimport (
    interval_t, initialize_interval_vector_t,
    make_interval_t_vector_with_size, make_interval_t_vector,
    interval_t_vector_append, free_interval_t_vector,
    interval_t_vector, interval_t_vector_reset, IntervalVector,
    interval_overlaps
)


cnp.import_array()

cdef long num_processors = PyInt_AsLong(cpu_count())


@cython.cdivision(True)
@cython.boundscheck(False)
cdef size_t binsearch(double* array, double x, size_t n) nogil:
    cdef:
        size_t lo, hi, mid
        double y, err

    lo = 0
    hi = n

    while hi != lo:
        mid = (hi + lo) / 2
        y = array[mid]
        err = y - x
        if hi - lo == 1:
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return 0


cdef struct spectrum_holder:
    double* mz
    double* intensity
    size_t size


cdef int prepare_arrays(list arrays, spectrum_holder** out) except 1:
    cdef:
        size_t i, n
        spectrum_holder* converted
        cnp.ndarray[double, ndim=1, mode='c'] mz
        cnp.ndarray[double, ndim=1, mode='c'] inten

    n = len(arrays)
    converted = <spectrum_holder*>malloc(sizeof(spectrum_holder) * n)
    if converted == NULL:
        PyErr_Format(MemoryError, "Failed to allocate array holder of size %lu", n)
        return 1
    for i in range(n):
        mz, inten = arrays[i]
        converted[i].size = mz.shape[0]
        if mz.shape[0]:
            converted[i].mz = &mz[0]
            converted[i].intensity = &inten[0]
        else:
            converted[i].mz = NULL
            converted[i].intensity = NULL
    out[0] = converted
    return 0


cdef double INF = float('inf')


cdef class GridAverager(object):
    cdef:
        public cnp.ndarray mz_axis
        public double min_mz
        public double max_mz
        public double dx
        double* mz_axis_
        public size_t size
        public size_t num_scans
        public cnp.ndarray intensities
        public cnp.ndarray empty
        interval_t_vector* occupied_indices


    def __init__(self, min_mz, max_mz, num_scans, dx=0.001, arrays=None):
        self.min_mz = min_mz
        self.max_mz = max_mz
        self.dx = dx
        self.num_scans = num_scans
        self.occupied_indices = NULL
        self.init_mz_axis()
        self.init_intensities_grid()

    cdef void init_mz_axis(self):
        cdef:
            cnp.ndarray[double, ndim=1, mode='c'] mz_axis
        self.mz_axis = mz_axis = np.arange(self.min_mz, self.max_mz + self.dx, self.dx)
        self.size = mz_axis.shape[0]
        self.mz_axis_ = &mz_axis[0]

    cdef int init_intensities_grid(self) except 1:
        cdef:
            size_t i
            int code

        self.intensities = np.zeros((self.num_scans, self.size), dtype=np.float64)
        self.empty = np.zeros(self.num_scans, dtype=np.uint8)
        self.occupied_indices = <interval_t_vector*>malloc(sizeof(interval_t_vector) * self.num_scans)
        if self.occupied_indices == NULL:
            PyErr_Format(MemoryError, "Failed to allocate index array of size %lu", self.num_scans)
            return 1
        for i in range(self.num_scans):
            code = initialize_interval_vector_t(&self.occupied_indices[i], 2)
            if code == 1:
                PyErr_Format(MemoryError, "Failed to allocate index array segment at %lu", i)
                return 1
        return 0

    cpdef release(self):
        cdef:
            size_t i

        if self.occupied_indices != NULL:
            for i in range(self.num_scans):
                free(self.occupied_indices[i].v)
            free(self.occupied_indices)
            self.occupied_indices = NULL


    def __dealloc__(self):
        self.release()

    cpdef IntervalVector get_occupied_indices(self, size_t i):
        if i > self.num_scans:
            raise IndexError(i)
        return IntervalVector.wrap(&self.occupied_indices[i])

    cpdef add_spectrum(self, cnp.ndarray[double, ndim=1, mode='c'] mz_array, cnp.ndarray[double, ndim=1, mode='c'] intensity_array, size_t index):
        self.empty[index] = mz_array.shape[0] == 0
        rebinned_intensities = self.create_intensity_axis(mz_array, intensity_array, index)
        return rebinned_intensities

    @cython.cdivision(True)
    cdef int _populate_intensity_axis(self, double* pmz, double* pinten, double* intensity_axis, size_t n, size_t m, interval_t_vector* occupied_acc) nogil:
        cdef:
            double x, mz_j, mz_j1, contrib, inten_j, inten_j1
            size_t i, j
            bint opened
            interval_t current_interval

        if m == 0:
            return 0
        current_interval.start = 0
        current_interval.end = 0
        opened = False
        for i in range(n):
            x = self.mz_axis_[i]
            j = binsearch(&pmz[0], x, m)
            mz_j = pmz[j]
            if mz_j <= x and j + 1 < m:
                mz_j1 = pmz[j + 1]
                inten_j = pinten[j]
                inten_j1 = pinten[j + 1]
            elif mz_j > x and j > 0:
                mz_j1 = mz_j
                inten_j1 = pinten[j]
                mz_j = pmz[j - 1]
                inten_j = pmz[j - 1]
            else:
                continue

            contrib = ((inten_j * (mz_j1 - x)) + (inten_j1 * (x - mz_j))) / (mz_j1 - mz_j)
            intensity_axis[i] = contrib
            if intensity_axis[i] > 0:
                if opened:
                    current_interval.end = i
                else:
                    current_interval.start = i
                    opened = True
            else:
                if opened:
                    current_interval.end = i
                    opened = False
                    if interval_t_vector_append(occupied_acc, current_interval) == -1:
                        return -1
        if opened:
            current_interval.end = i
            opened = False
            if interval_t_vector_append(occupied_acc, current_interval) == -1:
                return -1
        return 0

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cpdef cnp.ndarray create_intensity_axis(self, cnp.ndarray[double, ndim=1, mode='c'] mz_array, cnp.ndarray[double, ndim=1, mode='c'] intensity_array, size_t index):
        cdef:
            double* pmz
            double* pinten
            cnp.ndarray[double, ndim=1, mode='c'] intensity_axis
            double* intensity_axis_
            double x, mz_j, mz_j1, contrib, inten_j, inten_j1
            size_t i, j, n, m
            double[:, :] intensity_grid

        intensity_grid = self.intensities
        intensity_axis = intensity_grid[index, :]
        intensity_axis_ = &intensity_axis[0]

        m = mz_array.shape[0]
        if m == 0:
            return intensity_axis

        pmz = &mz_array[0]
        pinten = &intensity_array[0]
        n = self.size

        with nogil:
            interval_t_vector_reset(&self.occupied_indices[index])
            if self._populate_intensity_axis(pmz, pinten, intensity_axis_, n, m, &self.occupied_indices[index]) == -1:
                with gil:
                    raise ValueError("Failed to populate intensity axis")
        return intensity_axis

    @cython.boundscheck(False)
    cpdef add_spectra(self, list spectra, n_workers=None):
        cdef:
            spectrum_holder* spectrum_pairs
            spectrum_holder pair
            double[:, ::1] intensity_grid
            cnp.uint8_t[::1] emptiness
            double[:] intensity_frame
            int n_threads, i, n
            size_t start, stop

        if n_workers is None:
            n_workers = min(num_processors, len(spectra))
        n_threads = PyInt_AsLong(n_workers)


        prepare_arrays(spectra, &spectrum_pairs)
        n = PyList_GET_SIZE(spectra)

        intensity_grid = self.intensities
        emptiness = self.empty

        with nogil:
            for i in parallel.prange(n, num_threads=n_threads):
                pair = spectrum_pairs[i]
                emptiness[i] = pair.size == 0
                interval_t_vector_reset(&self.occupied_indices[i])
                if emptiness[i]:
                    continue
                if self._populate_intensity_axis(
                        pair.mz, pair.intensity, &intensity_grid[i, 0], self.size, pair.size, &self.occupied_indices[i]) == -1:
                    with gil:
                        raise ValueError("Failed to populate intensity axis")

            free(spectrum_pairs)


    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef cnp.ndarray[double, ndim=1, mode='c'] average_indices(self, size_t start, size_t stop, n_workers=None):
        cdef:
            size_t i, n, k
            interval_t_vector occupied
            interval_t current_interval
            cnp.ndarray[double, ndim=1, mode='c'] intensity_axis
            cnp.ndarray[double, ndim=1, mode='c'] intensity_frame
            double* intensity_frame_
            # double* intensity_axis_
            double[::1] intensity_axis_
            bint is_empty
            double normalizer
            cnp.uint8_t[::1] emptiness
            double[:, ::1] intensity_grid
            int n_threads
            ssize_t z
            long j, n_j

        if n_workers is None:
            n_threads = 1
        else:
            n_threads = PyInt_AsLong(n_workers)

        if start > stop:
            stop, start = start, stop
        if stop > self.num_scans:
            stop = self.num_scans

        n = PyList_GET_SIZE(self.intensities)

        intensity_axis = np.zeros_like(self.mz_axis)
        # intensity_axis_ = &intensity_axis[0]
        intensity_axis_ = intensity_axis
        normalizer = stop - start
        if normalizer == 0:
            normalizer = 1

        emptiness = self.empty
        intensity_grid = self.intensities

        z = self.size

        with nogil:
            for i in range(start, stop):
                is_empty = emptiness[i]
                if is_empty:
                    continue
                occupied = self.occupied_indices[i]
                n_j = occupied.used
                for j in parallel.prange(n_j, num_threads=n_threads):
                    current_interval = occupied.v[j]
                    for k in range(current_interval.start, current_interval.end):
                        intensity_axis_[k] += intensity_grid[i, k] / normalizer
        return intensity_axis

    cpdef get_occupied_intervals(self, size_t start, size_t stop):
        cdef:
            size_t i, n, k, head_i
            int advanced, did_update, keep_going
            interval_t_vector* occupied
            interval_t current_interval
            interval_t_vector* acc
            size_t* heads

        if stop < start:
            raise ValueError("Stop must be > Start")

        acc = make_interval_t_vector()
        k = stop - start
        heads = <size_t*>malloc(sizeof(size_t) * k)
        if heads == NULL:
            raise MemoryError()
        for i in range(k):
            heads[i] = 0

        has_more = True
        while has_more:
            current_interval.start = -1
            current_interval.end = -1

            for i in range(start, stop):
                occupied = &self.occupied_indices[i]
                head_i = heads[i - start]
                if occupied.used > 0 and head_i < occupied.used:
                    if occupied.v[head_i].start < current_interval.start:
                        current_interval.start = occupied.v[head_i].start
                    if occupied.v[head_i].end < current_interval.end:
                        current_interval.end = occupied.v[head_i].end + 1

            if current_interval.end < current_interval.start:
                current_interval.end = current_interval.start
            keep_going = True
            advanced = 0
            did_update = 0
            while keep_going:
                for i in range(start, stop):
                    occupied = &self.occupied_indices[i]
                    head_i = heads[i - start]
                    if head_i < occupied.used and interval_overlaps(&occupied.v[head_i], &current_interval):
                        current_interval.end = max(occupied.v[head_i].end, current_interval.end) + 1
                        heads[i - start] += 1
                        advanced += 1
                did_update += advanced
                keep_going = advanced
                advanced = 0

            if not did_update:
                has_more = False
            else:
                if acc.used > 0 and (acc.v[acc.used - 1].end == current_interval.start or interval_overlaps(&acc.v[acc.used - 1], &current_interval)):
                    acc.v[acc.used - 1].end = current_interval.end
                else:
                    interval_t_vector_append(acc, current_interval)


        out = IntervalVector.wrap(acc)
        out.set_should_free(True)
        # result = []
        # for i in range(acc.used):
        #     current_interval = acc.v[i]
        #     result.append((self.mz_axis[current_interval.start], self.mz_axis[current_interval.end]))
        # free_interval_t_vector(acc)
        free(heads)
        return out


    def __getitem__(self, i):
        if isinstance(i, slice):
            start = i.start or 0
            stop = i.stop or self.num_scans
        else:
            start = i - 1
            stop = i + 2
        if start < 0:
            start = 0
        if stop > self.num_scans:
            stop = self.num_scans
        return self.mz_axis, self.average_indices(start, stop)




@cython.cdivision(True)
@cython.boundscheck(False)
cpdef average_signal(object arrays, double dx=0.01, object weights=None, object num_threads=None):
    """Average multiple spectras' intensity arrays, with a common m/z axis

    Parameters
    ----------
    arrays : :class:`list` of pairs of :class:`np.ndarray`
        The m/z and intensity arrays to combine
    dx : float, optional
        The m/z resolution to build the averaged m/z axis with
    weights : :class:`list` of :class:`float`, optional
        Weight of each entry in `arrays`. Defaults to 1.0 for each if not provided.
    num_threads : int, optional
        The maximum number of threads to use while averaging signal. Defaults
        to the number of spectra being averaged or the maximum available from
        the hardware, whichever is smaller.

    Returns
    -------
    mz_array: :class:`np.ndarray`
    intensity_array: :class:`np.ndarray`
    """
    cdef:
        double lo, hi
        double mz_j, mz_j1
        double inten_j, inten_j1
        double x, contrib, tmp
        double scan_weight
        size_t i, j, n, n_arrays, n_points
        long k_array, n_scans, n_workers, worker_block
        object omz, ointen
        list convert
        int n_empty
        bint all_empty
        int error

        spectrum_holder* spectrum_pairs
        spectrum_holder pair

        cnp.ndarray[double, ndim=1, mode='c'] mz_array
        cnp.ndarray[double, ndim=1, mode='c'] intensity_array
        cnp.ndarray[double, ndim=1, mode='c'] mz
        cnp.ndarray[double, ndim=1, mode='c'] inten
        double* pweights
        double* pmz
        double* pinten
        double* intensity_array_local
        double[::1] pmz_array
        double[::1] pintensity_array_total
        double** intensity_layers

    if num_threads is None or num_threads < 0:
        n_workers = num_processors
    else:
        n_workers = num_threads
    if weights is None:
        weights = [1 for omz in arrays]
    elif len(arrays) != len(weights):
        raise ValueError("`arrays` and `weights` must have the same length")

    convert = []
    for omz, ointen in arrays:
        if len(omz) == 0:
            continue
        mz = omz.astype(np.double)
        inten = ointen.astype(np.double)
        convert.append((mz, inten))

    n_scans = PyList_GET_SIZE(convert)
    if n_scans == 0:
        return np.array([], dtype=np.double), np.array([], dtype=np.double)
    lo = INF
    hi = 0
    for mz, inten in convert:
        lo = min(mz[0], lo)
        hi = max(mz[mz.shape[0] - 1], hi)

    lo = max(lo - 1, 0)
    hi += 1
    pweights = <double*>malloc(sizeof(double) * n_scans)
    if pweights == NULL:
        printf("Unable to allocate memory for average_signal, could not allocate weights array of size %ld\n", n_scans)
        raise MemoryError("Unable to allocate memory for average_signal, could not allocate weights array of size %d" % (n_scans, ))
    for i in range(n_scans):
        pweights[i] = PyFloat_AsDouble(float(weights[i]))

    mz_array = np.arange(lo, hi, dx, dtype=np.double)
    intensity_array = np.zeros_like(mz_array, dtype=np.double)
    n_points = len(mz_array)
    pintensity_array_total = intensity_array
    pmz_array = mz_array
    error = 0
    intensity_layers = <double**>malloc(sizeof(double*) * n_scans)
    if intensity_layers == NULL:
        free(pweights)
        printf("Unable to allocate memory for average_signal, could not allocate temporary array of size %ld\n", n_scans)
        raise MemoryError("Unable to allocate memory for average_signal, could not allocate temporary array of size %d" % (n_scans, ))

    for k_array in range(n_scans):
        intensity_layers[k_array] = NULL

    for k_array in range(n_scans):
        intensity_layers[k_array] = <double*>malloc(sizeof(double) * n_points)
        if intensity_layers[k_array] == NULL:
            printf("Unable to allocate temporary array %d of size %zu for average_signal\n", k_array, n_points)
            error += 1

    if error:
        for k_array in range(n_scans):
            if intensity_layers[k_array] != NULL:
                free(intensity_layers[k_array])
        free(pweights)
        free(intensity_layers)
        printf("Unable to allocate memory for average_signal, failed %d partitions of size %zu\n", error, n_points)
        raise MemoryError("Unable to allocate memory for average_signal, failed %d partitions of size %d" % (error, n_points))

    prepare_arrays(convert, &spectrum_pairs)
    with nogil:
        if n_scans < n_workers:
            n_workers = n_scans
        for k_array in parallel.prange(n_scans, num_threads=n_workers):
            intensity_array_local = intensity_layers[k_array]
            pair = spectrum_pairs[k_array]
            if pair.size == 0:
                continue
            pmz = pair.mz
            pinten = pair.intensity
            contrib = 0
            scan_weight = pweights[k_array]
            for i in range(n_points):
                x = pmz_array[i]
                j = binsearch(pmz, x, pair.size)
                mz_j = pmz[j]
                if mz_j <= x and j + 1 < pair.size:
                    mz_j1 = pmz[j + 1]
                    inten_j = pinten[j]
                    inten_j1 = pinten[j + 1]
                elif mz_j > x and j > 0:
                    mz_j1 = mz_j
                    inten_j1 = pinten[j]
                    mz_j = pmz[j - 1]
                    inten_j = pmz[j - 1]
                else:
                    continue
                tmp = mz_j1 - mz_j
                if tmp == 0:
                    contrib = 0.0
                else:
                    contrib = ((inten_j * (mz_j1 - x)) + (inten_j1 * (x - mz_j))) / (mz_j1 - mz_j)
                intensity_array_local[i] = contrib * scan_weight

        for k_array in range(n_scans):
            intensity_array_local = intensity_layers[k_array]
            if intensity_array_local == NULL:
                continue
            for i in range(n_points):
                pintensity_array_total[i] += intensity_array_local[i]
            free(intensity_array_local)
        free(intensity_layers)
        scan_weight = 0
        for i in range(n_scans):
            scan_weight += pweights[i]
        for i in range(n_points):
            pintensity_array_total[i] /= scan_weight
        free(spectrum_pairs)
        free(pweights)
    if error:
        raise MemoryError("Unable to allocate memory for average_signal, failed %d partitions of size %d" % (error, n_points))
    return mz_array, intensity_array



cimport cython

from libc.stdlib cimport malloc, realloc, free
from libc cimport *

from cpython.exc cimport PyErr_BadArgument
from cpython.mem cimport PyObject_Malloc, PyObject_Free

from cpython.sequence cimport (
    PySequence_Size, PySequence_Check, PySequence_Fast,
    PySequence_Fast_GET_ITEM, PySequence_Fast_GET_SIZE)
from cpython.slice cimport PySlice_GetIndicesEx

from cpython.int cimport PyInt_FromSize_t, PyInt_AsLong

cdef enum VectorStateEnum:
    should_free = 1


cdef int compare_value_size_t(const void* a, const void* b) nogil:
    cdef:
        size_t av, bv
    av = (<size_t*>a)[0]
    bv = (<size_t*>b)[0]
    if av < bv:
        return -1
    elif av == bv:
        return 0
    else:
        return 1


cdef int compare_value_size_t_reverse(const void* a, const void* b) nogil:
    return -compare_value_size_t(a, b)



cdef extern from * nogil:
    int printf (const char *template, ...)
    void qsort (void *base, unsigned short n, unsigned short w, int (*cmp_func)(void*, void*))


DEF GROWTH_RATE = 2
DEF INITIAL_SIZE = 4


cdef size_t_vector* make_size_t_vector_with_size(size_t size) nogil:
    cdef:
        size_t_vector* vec

    vec = <size_t_vector*>malloc(sizeof(size_t_vector))
    vec.v = <size_t*>malloc(sizeof(size_t) * size)
    vec.size = size
    vec.used = 0

    return vec


cdef size_t_vector* make_size_t_vector() nogil:
    return make_size_t_vector_with_size(INITIAL_SIZE)


cdef int size_t_vector_resize(size_t_vector* vec) nogil:
    cdef:
        size_t new_size
        size_t* v
    new_size = vec.size * GROWTH_RATE
    v = <size_t*>realloc(vec.v, sizeof(size_t) * new_size)
    if v == NULL:
        printf("size_t_vector_resize returned -1\n")
        return -1
    vec.v = v
    vec.size = new_size
    return 0


cdef int size_t_vector_append(size_t_vector* vec, size_t value) nogil:
    if (vec.used + 1) >= vec.size:
        if size_t_vector_resize(vec) == -1:
            return -1
    vec.v[vec.used] = value
    vec.used += 1
    return 0


cdef void free_size_t_vector(size_t_vector* vec) nogil:
    free(vec.v)
    free(vec)


cdef void print_size_t_vector(size_t_vector* vec) nogil:
    cdef:
        size_t i
    i = 0
    printf("[")
    while i < vec.used:
        printf("%0.6f", vec.v[i])
        if i != (vec.used - 1):
            printf(", ")
        i += 1
    printf("]\n")


cdef void size_t_vector_reset(size_t_vector* vec) nogil:
    vec.used = 0


cdef int size_t_vector_reserve(size_t_vector* vec, size_t new_size) nogil:
    cdef:
        size_t* v
    v = <size_t*>realloc(vec.v, sizeof(size_t) * new_size)
    if v == NULL:
        printf("size_t_vector_resize returned -1\n")
        return -1
    vec.v = v
    vec.size = new_size
    if new_size > vec.used:
        vec.used = new_size
    return 0



cdef char* SizeTVector_buffer_type_code = "Q"


@cython.final
@cython.freelist(512)
cdef class SizeTVector(object):

    @staticmethod
    cdef SizeTVector _create(size_t size):
        cdef:
            SizeTVector self
        self = SizeTVector.__new__(SizeTVector)
        self.flags = 0
        self.allocate_storage_with_size(size)
        return self

    @staticmethod
    cdef SizeTVector wrap(size_t_vector* vector):
        cdef:
            SizeTVector self
        self = SizeTVector.__new__(SizeTVector)
        self.flags = 0
        self.impl = vector
        self.set_should_free(False)
        return self

    def __init__(self, seed=None):
        cdef:
            size_t n
        self.flags = 0
        if seed is not None:
            if not PySequence_Check(seed):
                raise TypeError("Must provide a Sequence-like object")
            n = len(seed)
            self.allocate_storage_with_size(n)
            self.extend(seed)
        else:
            self.allocate_storage()

    cdef int allocate_storage_with_size(self, size_t size) nogil:
        if self.impl != NULL:
            if self.flags & VectorStateEnum.should_free:
                free_size_t_vector(self.impl)
        self.impl = make_size_t_vector_with_size(size)
        self.flags |= VectorStateEnum.should_free
        return self.impl == NULL

    cdef int allocate_storage(self) nogil:
        if self.impl != NULL:
            if self.flags & VectorStateEnum.should_free:
                free_size_t_vector(self.impl)
        self.impl = make_size_t_vector()
        self.flags |= VectorStateEnum.should_free
        return self.impl == NULL

    cdef int free_storage(self) nogil:
        free_size_t_vector(self.impl)

    cdef bint get_should_free(self) nogil:
        return self.flags & VectorStateEnum.should_free

    cdef void set_should_free(self, bint flag) nogil:
        self.flags &= VectorStateEnum.should_free * flag

    cdef size_t* get_data(self) nogil:
        return self.impl.v

    cdef size_t get(self, size_t i) nogil:
        return self.impl.v[i]

    cdef void set(self, size_t i, size_t value) nogil:
        self.impl.v[i] = value

    cdef size_t size(self) nogil:
        return self.impl.used

    cdef int cappend(self, size_t value) nogil:
        return size_t_vector_append(self.impl, value)

    cpdef int append(self, object value) except *:
        cdef:
            size_t cvalue
        cvalue = self._to_c(value)
        return self.cappend(cvalue)

    cpdef int extend(self, object values) except *:
        cdef:
            size_t i, n
            object fast_seq
        if not PySequence_Check(values):
            raise TypeError("Must provide a Sequence-like object")

        fast_seq = PySequence_Fast(values, "Must provide a Sequence-like object")
        n = PySequence_Fast_GET_SIZE(values)

        for i in range(n):
            if self.append(<object>PySequence_Fast_GET_ITEM(fast_seq, i)) != 0:
                return 1

    cpdef int reserve(self, size_t size) nogil:
        return size_t_vector_reserve(self.impl, size)

    cpdef int fill(self, size_t value) nogil:
        cdef:
            size_t i, n
        n = self.size()
        for i in range(n):
            self.set(i, value)
        return 0

    cpdef SizeTVector copy(self):
        cdef:
            SizeTVector dup
            size_t i, n
        n = self.size()
        dup = SizeTVector._create(n)
        for i in range(n):
            dup.cappend(self.get(i))
        return dup

    cdef SizeTVector _slice(self, object slice_spec):
        cdef:
            SizeTVector dup
            Py_ssize_t length, start, stop, step, slice_length, i
        PySlice_GetIndicesEx(
            slice_spec, self.size(), &start, &stop, &step, &slice_length)
        dup = SizeTVector._create(slice_length)
        i = start
        while i < stop:
            dup.cappend(self.get(i))
            i += step
        return dup

    def __dealloc__(self):
        if self.get_should_free():
            self.free_storage()

    def __len__(self):
        return self.size()

    def __iter__(self):
        cdef:
            size_t i, n
        n = self.size()
        for i in range(n):
            yield self._to_python(self.get(i))

    def __getitem__(self, i):
        cdef:
            Py_ssize_t index
            size_t n
        if isinstance(i, slice):
            return self._slice(i)
        index = i
        n = self.size()
        if index < 0:
            index = n + index
        if index > n or index < 0:
            raise IndexError(index)
        return self._to_python(self.get(index))

    def __setitem__(self, i, value):
        cdef:
            Py_ssize_t index
            size_t n
        if isinstance(i, slice):
            raise TypeError("Does not support slice-assignment yet")
        n = self.size()
        index = i
        if index < 0:
            index = n + index
        if index > self.size() or index < 0:
            raise IndexError(i)
        self.set(index, self._to_c(value))

    def __repr__(self):
        return "{self.__class__.__name__}({members})".format(self=self, members=list(self))


    def __getbuffer__(self, Py_buffer* info, int flags):
        # This implementation of getbuffer is geared towards Cython
        # requirements, and does not yet fulfill the PEP.
        # In particular strided access is always provided regardless
        # of flags
        cdef size_t item_count = self.size()

        info.suboffsets = NULL
        info.buf = <char*>self.impl.v
        info.readonly = 0
        info.ndim = 1
        info.itemsize = sizeof(size_t)
        info.len = info.itemsize * item_count

        info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
        if not info.shape:
            raise MemoryError()
        info.shape[0] = item_count      # constant regardless of resizing
        info.strides = &info.itemsize

        info.format = SizeTVector_buffer_type_code
        info.obj = self

    def __releasebuffer__(self, Py_buffer* info):
        PyObject_Free(info.shape)


    cpdef void qsort(self, bint reverse=False) nogil:
        if reverse:
            qsort(self.get_data(), self.size(), sizeof(size_t), compare_value_size_t_reverse)
        else:
            qsort(self.get_data(), self.size(), sizeof(size_t), compare_value_size_t)

    cpdef object _to_python(self, size_t value):
        return PyInt_FromSize_t(value)

    cpdef size_t _to_c(self, object value) except *:
        return PyInt_AsLong(value)


cdef struct DoubleVector:
    double* v
    size_t used
    size_t size

cdef DoubleVector* make_double_vector_with_size(size_t size) nogil
cdef DoubleVector* make_double_vector() nogil
cdef int double_vector_resize(DoubleVector* vec) nogil
cdef int double_vector_append(DoubleVector* vec, double value) nogil
cdef void free_double_vector(DoubleVector* vec) nogil
cdef void print_double_vector(DoubleVector* vec) nogil
cdef void reset_double_vector(DoubleVector* vec) nogil
cdef list double_vector_to_list(DoubleVector* vec)
cdef DoubleVector* list_to_double_vector(list input_list)


from libc.stdlib cimport malloc, realloc, free
from libc cimport *

cdef extern from * nogil:
    int printf (const char *template, ...)


cdef DoubleVector* make_double_vector_with_size(size_t size) nogil:
    cdef:
        DoubleVector* vec

    vec = <DoubleVector*>malloc(sizeof(DoubleVector))
    vec.v = <double*>malloc(sizeof(double) * size)
    vec.size = size
    vec.used = 0

    return vec


cdef DoubleVector* make_double_vector() nogil:
    return make_double_vector_with_size(4)


cdef int double_vector_resize(DoubleVector* vec) nogil:
    cdef:
        size_t new_size
        double* v
    new_size = vec.size * 2
    v = <double*>realloc(vec.v, sizeof(double) * new_size)
    if v == NULL:
        printf("double_vector_resize returned -1\n")
        return -1
    vec.v = v
    vec.size = new_size
    return 0


cdef int double_vector_append(DoubleVector* vec, double value) nogil:
    if (vec.used + 1) == vec.size:
        double_vector_resize(vec)
    vec.v[vec.used] = value
    vec.used += 1
    return 0


cdef void free_double_vector(DoubleVector* vec) nogil:
    free(vec.v)
    free(vec)


cdef void print_double_vector(DoubleVector* vec) nogil:
    cdef:
        size_t i
    i = 0
    printf("[")
    while i < vec.used:
        printf("%0.6f", vec.v[i])
        if i != (vec.used - 1):
            printf(", ")
        i += 1
    printf("]\n")


cdef void reset_double_vector(DoubleVector* vec) nogil:
    vec.used = 0


cdef list double_vector_to_list(DoubleVector* vec):
    cdef:
        size_t i
        list result
    result = list()
    i = 0
    while i < vec.used:
        result.append(vec.v[i])
        i += 1
    return result


cdef DoubleVector* list_to_double_vector(list input_list):
    cdef:
        DoubleVector* vec
        double val
    vec = make_double_vector_with_size(len(input_list))
    for val in input_list:
        double_vector_append(vec, val)
    return vec


def test():
    cdef list listy
    cdef DoubleVector* vecty

    listy = [1., 2., 3., 4., 23121.]
    vecty = list_to_double_vector(listy)
    print_double_vector(vecty)
    listy = []
    print(listy)
    listy = double_vector_to_list(vecty)
    free_double_vector(vecty)
    print(listy)
    del listy



import os
from .peak_picker import PeakProcessor, pick_peaks, fit_type_map, peak_mode_map
from .peak_set import PeakSet, FittedPeak, is_peak, simple_peak
from .peak_index import PeakIndex
from . import peak_statistics
from . import search
from . import fticr_denoising
from . import fft_patterson_charge_state
from . import scan_filter
from .base import PeakLike
from .reprofile import reprofile
from .scan_averaging import average_signal, GridAverager
from .smoothing import gaussian_smooth
from .version import version as __version__

__all__ = ["PeakProcessor", "pick_peaks", "PeakIndex", "PeakSet",
           "FittedPeak", "peak_statistics", "search", "fticr_denoising",
           "scan_filter", "fft_patterson_charge_state", "get_include",
           "fit_type_map", "peak_mode_map", "reprofile", "average_signal",
           "gaussian_smooth", 'is_peak', 'simple_peak', "PeakLike", "GridAverager"]


def get_include():
    """Retrieve the path to compiled C extensions' source files to make linking simple.
    """
    return os.path.join(__path__[0], "_c")


try:
    from ms_peak_picker._c import peak_picker as _cpeak_picker
    has_c = True
    _has_c_error = None
except ImportError as _has_c_error:
    has_c = False


def check_c_extensions():
    if has_c:
        print("C extensions appear to have imported successfully.")
        return True
    else:
        print("Could not import peak picking machinery: %r" % (_has_c_error, ))
    return False



'''A collection of tools for drawing and annotating mass spectra
'''
from collections import namedtuple

from matplotlib import pyplot as plt
import numpy as np


from .base import PeakLike


point = namedtuple('point', ('mz', 'intensity'))


def draw_raw(mz_array, intensity_array=None, ax=None, normalize=False, **kwargs):
    """Draws un-centroided profile data, visualizing continuous
    data points

    Parameters
    ----------
    mz_array : :class:`np.ndarray` or :class:`tuple`
        Either the m/z array to be visualized, or if `intensity_array`
        is `None`, `mz_array` will be unpacked, looking to find a sequence
        of two `np.ndarray` objects for the m/z (X) and intensity (Y)
        coordinates
    intensity_array : :class:`np.ndarray`, optional
        The intensity array to be visualized. If `None`, will attempt to
        unpack `mz_array`
    ax : :class:`matplotlib.Axes`, optional
        The axis to draw the plot on. If missing, a new one will be created using
        :func:`matplotlib.pyplot.subplots`
    normalize: bool, optional
        if `True`, will normalize the abundance dimension to be between 0 and 100%
    pretty: bool, optional
        If `True`, will call :func:`_beautify_axes` on `ax`
    **kwargs
        Passed to :meth:`matplotlib.Axes.plot`

    Returns
    -------
    :class:`~.Axes`
    """
    pretty = kwargs.pop("pretty", True)
    if intensity_array is None and len(mz_array) == 2:
        mz_array, intensity_array = mz_array
    if ax is None:
        _, ax = plt.subplots(1)
    if normalize:
        intensity_array = intensity_array / intensity_array.max() * 100.0
    ax.plot(mz_array, intensity_array, **kwargs)
    ax.set_xlabel("m/z")
    ax.set_ylabel("Relative Intensity")
    if pretty:
        if intensity_array.shape[0] > 0:
            set_ylim = intensity_array.min() >= 0
        else:
            set_ylim = True
        _beautify_axes(ax, set_ylim, normalize)
    return ax


def peaklist_to_vector(peaklist, width=0.000001):
    """Convert a list of discrete centroided peaks into a pair of continuous m/z
    and intensity arrays

    Parameters
    ----------
    peaklist : :class:`~Iterable` of :class:`~.PeakLike`
        The collection of peaks to convert
    width : float, optional
        The spacing between the center of the peak and it's shoulders

    Returns
    -------
    np.ndarray:
        The generated m/z array
    np.ndarray:
        The generated intensity array

    Raises
    ------
    TypeError
        When the input could not be coerced into a peak list
    """
    try:
        mzs = []
        intensities = []
        for peak in sorted(peaklist, key=lambda x: x.mz):
            mzs.append(peak.mz - width)
            intensities.append(0.)
            mzs.append(peak.mz)
            intensities.append(peak.intensity)
            mzs.append(peak.mz + width)
            intensities.append(0.)
        return np.array(mzs), np.array(intensities)
    except AttributeError:
        pt = peaklist[0]
        if not PeakLike.is_a(pt):
            try:
                if len(pt) == 2:
                    peaklist = [point(*p) for p in peaklist]
                    return peaklist_to_vector(peaklist, width)
            except Exception:
                pass
        raise TypeError("Expected a sequence of peak-like objects"
                        " or (mz, intensity) pairs, but got %r instead" % type(pt))


def draw_peaklist(peaklist, ax=None, normalize=False, **kwargs):
    """Draws centroided peak data, visualizing peak apexes.

    The peaks will be converted into a single smooth curve using
    :func:`peaklist_to_vector`.

    Parameters
    ----------
    peaklist: :class:`Iterable` of :class:`~.PeakLike`
        The peaks to draw.
    ax : matplotlib.Axes, optional
        The axis to draw the plot on. If missing, a new one will be created using
        :func:`matplotlib.pyplot.subplots`
    normalize: bool, optional
        if `True`, will normalize the abundance dimension to be between 0 and 100%
    pretty: bool, optional
        If `True`, will call :func:`_beautify_axes` on `ax`
    **kwargs
        Passed to :meth:`matplotlib.Axes.plot`

    Returns
    -------
    matplotlib.Axes
    """
    pretty = kwargs.pop("pretty", True)
    if ax is None:
        _, ax = plt.subplots(1)
    mz_array, intensity_array = peaklist_to_vector(peaklist)
    if normalize:
        intensity_array = intensity_array / intensity_array.max() * 100.0
    ax.plot(mz_array, intensity_array, **kwargs)
    ax.set_xlabel("m/z")
    ax.set_ylabel("Relative Intensity")
    if pretty:
        if intensity_array.shape[0] > 0:
            set_ylim = intensity_array.min() >= 0
        else:
            set_ylim = True
        _beautify_axes(ax, set_ylim, normalize)
    return ax


def mirror_peaks(peaklist1, peaklist2, ax=None, normalize=True, kwargs1=None, kwargs2=None, **kwargs):
    """Create a mirror plot of two peak lists.

    Plot one peak list upside-down below another peak list. Commonly used to compare/contrast two
    two peak lists visually.

    Parameters
    ----------
    peaklist1 : :class:`~.Iterable`
        The peak list to plot right-side up.
    peaklist2 : :class:`~.Iterable`
        The peak list to plot upside-down.
    ax : matplotlib.Axes, optional
        The axis to draw the plot on. If missing, a new one will be created using
        :func:`matplotlib.pyplot.subplots`
    normalize : bool, optional
        Whether or not to normalize the intensity scales of the spectra.
        The default is :const:`True`, unlike other drawing functions, in
        order to keep the scales of the spectra relatively equal.
    kwargs1 : dict, optional
        Any keyword arguments to be applied solely when drawing `peaklist1`.
    kwargs2 : dict, optional
        Any keyword arguments to be applied solely when drawing `peaklist2`.
    pretty: bool, optional
        If `True`, will call :func:`_beautify_axes` on `ax`
    **kwargs:
        Keyword arguments to be used when drawing both peak lists.
        Passed to :meth:`matplotlib.Axes.plot`

    Returns
    -------
    :class:`matplotlib.Axes`
    """
    pretty = kwargs.pop("pretty", True)
    if kwargs1 is None:
        kwargs1 = {}
    if kwargs2 is None:
        kwargs2 = {}
    if ax is None:
        _, ax = plt.subplots(1)

    # Draw first peak list
    mz_array, intensity_array = peaklist_to_vector(peaklist1)
    if normalize:
        intensity_array = intensity_array / intensity_array.max() * 100.0
    kwargs1.update(kwargs)
    ax.plot(mz_array, intensity_array, **kwargs1)

    # Draw second peak list
    mz_array, intensity_array = peaklist_to_vector(peaklist2)
    if normalize:
        intensity_array = intensity_array / intensity_array.max() * 100.0
    intensity_array = -1 * intensity_array
    kwargs2.update(kwargs)
    ax.plot(mz_array, intensity_array, **kwargs2)

    # Update the axes if needed.
    if pretty:
        _beautify_axes(ax, False)
        _normalize_ylabels(ax, adjust_sign=True)
    return ax


def _beautify_axes(ax, set_ylim=True, normalize=False):
    ax.axes.spines['right'].set_visible(False)
    ax.axes.spines['top'].set_visible(False)
    ax.yaxis.tick_left()
    ax.xaxis.tick_bottom()
    ax.xaxis.set_ticks_position('none')
    ax.get_xaxis().get_major_formatter().set_useOffset(False)
    if set_ylim:
        ax.set_ylim(0, max(ax.get_ylim()))
    if normalize:
        _normalize_ylabels(ax)
    return ax


def _normalize_ylabels(ax, adjust_sign=True):
    ytick_labels = ax.get_yticklabels()
    yticks = ax.get_yticks()
    for tl, tick in zip(ytick_labels, yticks):
        txt = tl.get_text()
        if txt == "":
            txt = str(tick)
        if adjust_sign:
            txt = txt.replace("-", "")
        txt += "%"
        tl.set_text(txt)
    ax.set_yticklabels(ytick_labels)
    return ax


__all__ = [
    "draw_peaklist", "draw_raw"
]



import sys
import traceback
import os
import platform

from setuptools import setup, Extension as _Extension, find_packages

from distutils.command.build_ext import build_ext
from distutils.errors import (CCompilerError, DistutilsExecError,
                              DistutilsPlatformError)


def has_option(name):
    try:
        sys.argv.remove('--%s' % name)
        return True
    except ValueError:
        pass
    # allow passing all cmd line options also as environment variables
    env_val = os.getenv(name.upper().replace('-', '_'), 'false').lower()
    if env_val == "true":
        return True
    return False


# with_openmp = has_option('with-openmp')
no_openmp = has_option('no-openmp')
debug_symbols = has_option("debug")

with_openmp = not no_openmp

include_diagnostics = has_option("include-diagnostics")
force_cythonize = has_option("force-cythonize")

use_python_implementation = has_option("pure-python")

print("Building with OpenMP? %s" % with_openmp)


def configure_openmp(ext):
    # http://www.microsoft.com/en-us/download/confirmation.aspx?id=2092 was required.
    if os.name == 'nt' and with_openmp:
        ext.extra_compile_args.append("/openmp")
    elif platform.system() == 'Darwin':
        pass
    elif with_openmp:
        ext.extra_compile_args.append("-fopenmp")
        ext.extra_link_args.append("-fopenmp")


def Extension(*args, **kwargs):
    ext = _Extension(*args, **kwargs)
    if debug_symbols:
        if os.name == 'nt':
            ext.extra_compile_args.append("-Zi")
            ext.extra_compile_args.append("-Ox")
            ext.extra_link_args.append("-debug:full")
        else:
            ext.extra_compile_args.append("-g3")
            ext.extra_compile_args.append("-O0")
    # ext.extra_compile_args.append("/fsanitize=address")
    return ext

def OpenMPExtension(*args, **kwargs):
    ext = Extension(*args, **kwargs)
    configure_openmp(ext)
    return ext


def make_cextensions():
    import numpy
    macros = []
    try:
        from Cython.Build import cythonize
        cython_directives = {
            'embedsignature': True,
            "profile": include_diagnostics
        }
        if include_diagnostics:
            macros.append(("CYTHON_TRACE_NOGIL", "1"))
        print("Using Directives", cython_directives)
        extensions = cythonize([
            OpenMPExtension(
                name="ms_peak_picker._c.peak_statistics", sources=["src/ms_peak_picker/_c/peak_statistics.pyx"],
                include_dirs=[numpy.get_include()], define_macros=macros),
            Extension(name='ms_peak_picker._c.peak_set', sources=["src/ms_peak_picker/_c/peak_set.pyx"]),
            Extension(name='ms_peak_picker._c.fft_patterson_charge_state',
                      sources=["src/ms_peak_picker/_c/fft_patterson_charge_state.pyx"],
                      include_dirs=[numpy.get_include()], define_macros=macros),
            Extension(name="ms_peak_picker._c.search", sources=["src/ms_peak_picker/_c/search.pyx"],
                      include_dirs=[numpy.get_include()], define_macros=macros),
            Extension(name="ms_peak_picker._c.peak_index", sources=["src/ms_peak_picker/_c/peak_index.pyx"],
                      include_dirs=[numpy.get_include()], define_macros=macros),
            Extension(name='ms_peak_picker._c.double_vector', sources=["src/ms_peak_picker/_c/double_vector.pyx"]),
            Extension(name='ms_peak_picker._c.size_t_vector', sources=[
                      "src/ms_peak_picker/_c/size_t_vector.pyx"]),
            Extension(name='ms_peak_picker._c.interval_t_vector', sources=[
                      "src/ms_peak_picker/_c/interval_t_vector.pyx"]),
            Extension(name='ms_peak_picker._c.smoother', sources=["src/ms_peak_picker/_c/smoother.pyx"],
                      include_dirs=[numpy.get_include()], define_macros=macros),
            OpenMPExtension(name='ms_peak_picker._c.scan_averaging',
                            sources=['src/ms_peak_picker/_c/scan_averaging.pyx'],
                            include_dirs=[numpy.get_include()], define_macros=macros),
            Extension(name='ms_peak_picker._c.fticr_denoising', sources=['src/ms_peak_picker/_c/fticr_denoising.pyx'],
                      include_dirs=[numpy.get_include()], define_macros=macros),
            Extension(name="ms_peak_picker._c.peak_picker", sources=['src/ms_peak_picker/_c/peak_picker.pyx'],
                      include_dirs=[numpy.get_include()], define_macros=macros),
        ], compiler_directives=cython_directives, force=force_cythonize)
    except ImportError:
        extensions = ([
            OpenMPExtension(
                name="ms_peak_picker._c.peak_statistics", sources=["src/ms_peak_picker/_c/peak_statistics.c"],
                include_dirs=[numpy.get_include()], define_macros=macros),
            Extension(name='ms_peak_picker._c.peak_set', sources=[
                      "src/ms_peak_picker/_c/peak_set.c"], define_macros=macros),
            Extension(name='ms_peak_picker._c.fft_patterson_charge_state',
                      sources=["src/ms_peak_picker/_c/fft_patterson_charge_state.c"],
                      include_dirs=[numpy.get_include()], define_macros=macros),
            Extension(name="ms_peak_picker._c.search", sources=["src/ms_peak_picker/_c/search.c"],
                      include_dirs=[numpy.get_include()], define_macros=macros),
            Extension(name="ms_peak_picker._c.peak_index", sources=["src/ms_peak_picker/_c/peak_index.c"],
                      include_dirs=[numpy.get_include()], define_macros=macros),
            Extension(name='ms_peak_picker._c.double_vector', sources=[
                      "src/ms_peak_picker/_c/double_vector.c"], define_macros=macros),
            Extension(name='ms_peak_picker._c.size_t_vector', sources=[
                "src/ms_peak_picker/_c/size_t_vector.c"]),
            Extension(name='ms_peak_picker._c.interval_t_vector', sources=[
                      "src/ms_peak_picker/_c/interval_t_vector.c"]),
            Extension(name='ms_peak_picker._c.smoother', sources=["src/ms_peak_picker/_c/smoother.c"],
                      include_dirs=[numpy.get_include()], define_macros=macros),
            OpenMPExtension(name='ms_peak_picker._c.scan_averaging',
                            sources=['src/ms_peak_picker/_c/scan_averaging.c'],
                            include_dirs=[numpy.get_include()], define_macros=macros),
            Extension(name='ms_peak_picker._c.fticr_denoising', sources=['src/ms_peak_picker/_c/fticr_denoising.c'],
                      include_dirs=[numpy.get_include()], define_macros=macros)
        ])
    return extensions


ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)
if sys.platform == 'win32':
    # 2.6's distutils.msvc9compiler can raise an IOError when failing to
    # find the compiler
    ext_errors += (IOError,)


class BuildFailed(Exception):

    def __init__(self):
        self.cause = sys.exc_info()[1]  # work around py 2/3 different syntax

    def __str__(self):
        return str(self.cause)


class ve_build_ext(build_ext):
    # This class allows C extension building to fail.

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            traceback.print_exc()
            raise BuildFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except ext_errors:
            traceback.print_exc()
            raise BuildFailed()
        except ValueError:
            # this can happen on Windows 64 bit, see Python issue 7511
            traceback.print_exc()
            if "'path'" in str(sys.exc_info()[1]):  # works with both py 2/3
                raise BuildFailed()
            raise


cmdclass = {}

cmdclass['build_ext'] = ve_build_ext


def status_msgs(*msgs):
    print('*' * 75)
    for msg in msgs:
        print(msg)
    print('*' * 75)


with open("src/ms_peak_picker/version.py") as version_file:
    version = None
    for line in version_file.readlines():
        if "version = " in line:
            version = line.split(" = ")[1].replace("\"", "").strip()
            print("Version is: %r" % (version,))
            break
    else:
        print("Cannot determine version")


def run_setup(include_cext=True):
    setup(
        name='ms_peak_picker',
        description='A library to pick peaks from mass spectral data',
        long_description='A library to pick peaks from mass spectral data',
        version=version,
        packages=find_packages(where='src'),
        package_dir={"": "src"},
        zip_safe=False,
        install_requires=['numpy', "scipy", "six"],
        ext_modules=make_cextensions() if include_cext else None,
        cmdclass=cmdclass,
        author='Joshua Klein',
        author_email="jaklein@bu.edu",
        maintainer='Joshua Klein',
        maintainer_email="jaklein@bu.edu",
        include_package_data=True,
        classifiers=[
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Topic :: Education',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Topic :: Scientific/Engineering :: Chemistry',
            'Topic :: Software Development :: Libraries'
        ],
        license='License :: OSI Approved :: Apache Software License')


try:
    run_setup(True)
except Exception as exc:
    import traceback
    traceback.print_exc()
    if use_python_implementation:
        run_setup(False)
        status_msgs(
            "WARNING: The C extension could not be compiled, " +
            "speedups are not enabled.",
            "Plain-Python build succeeded."
        )
    else:
        raise



cimport cython
cimport numpy as np

from ms_peak_picker._c.peak_set cimport FittedPeak

ctypedef np.float64_t DTYPE_t


cpdef cython.floating find_signal_to_noise(
    double target_val,
    np.ndarray[cython.floating, ndim=1, mode='c'] intensity_array,
    size_t index)

cpdef DTYPE_t find_full_width_at_half_max(np.ndarray[cython.floating, ndim=1, mode='c'] mz_array,
                                          np.ndarray[cython.floating, ndim=1, mode='c'] intensity_array,
                                          size_t data_index,
                                          double signal_to_noise=*)

cpdef DTYPE_t find_right_width(np.ndarray[cython.floating, ndim=1, mode='c'] mz_array,
                               np.ndarray[cython.floating, ndim=1, mode='c'] intensity_array,
                               size_t data_index, DTYPE_t signal_to_noise=*)

cpdef DTYPE_t find_left_width(np.ndarray[cython.floating, ndim=1, mode='c'] mz_array,
                              np.ndarray[cython.floating, ndim=1, mode='c'] intensity_array,
                              size_t data_index, DTYPE_t signal_to_noise=*)

cpdef cython.floating quadratic_fit(np.ndarray[cython.floating, ndim=1, mode='c'] mz_array,
                            np.ndarray[cython.floating, ndim=1, mode='c'] intensity_array,
                            ssize_t index)

cpdef double lorentzian_fit(np.ndarray[cython.floating, ndim=1, mode='c'] mz_array,
                            np.ndarray[cython.floating, ndim=1, mode='c'] intensity_array,
                            size_t index, double full_width_at_half_max)

cpdef double peak_area(np.ndarray[cython.floating, ndim=1, mode='c'] mz_array,
                       np.ndarray[cython.floating, ndim=1, mode='c'] intensity_array,
                       size_t start, size_t stop)


cdef class PeakShapeModel(object):
    cdef:
        public FittedPeak peak
        public double center

    cpdef double predict(self, double mz)

    cpdef object shape(self)

    cpdef double volume(self)

    cpdef double error(self, double mz, double intensity)


cdef class GaussianModel(PeakShapeModel):
    pass


cdef class PeakSetReprofiler(object):
    cdef:
        public list models
        public np.ndarray gridx
        public np.ndarray gridy
        public double dx

    cpdef size_t _find_starting_model(self, double x)

    cdef void _build_grid(self)

    cpdef _reprofile(self)



cimport cython
cimport numpy as np
from libc cimport math
import numpy as np

from scipy import interpolate

ctypedef np.float64_t DTYPE_t


@cython.nonecheck(False)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef np.ndarray[DTYPE_t, ndim=1] autocorr(np.ndarray[DTYPE_t, ndim=1] inp):
    cdef:
        np.ndarray[DTYPE_t, ndim=1] out
        size_t n, j, i, top
        DTYPE_t average, normalizer, total

    out = np.zeros_like(inp)
    average = inp.mean()
    n = len(inp)
    normalizer = <DTYPE_t>n

    for i in range(n):
        total = 0.0
        top = n - i - 1

        for j in range(top):
            total += (inp[j] - average) * (inp[j + i] - average)
            if j > 0:
                out[i] = (total / normalizer)
            else:
                out[i] = 0.
    return out


@cython.nonecheck(False)
@cython.cdivision(True)
@cython.boundscheck(False)
cdef (double, int) _best_scoring_charge_state(double min_mz, double max_mz, size_t start_index, list autocorr_scores, int max_charge):
    cdef:
        double best_autoscore, current_score
        bint going_up, was_going_up
        size_t npoints, i
        int best_charge_state, charge_state
    best_autoscore = -1
    best_charge_state = -1

    going_up = False
    was_going_up = False

    npoints = len(autocorr_scores)
    for i in range(start_index, npoints):
        if i < 2:
            continue
        going_up = (autocorr_scores[i] - autocorr_scores[i - 1]) > 0
        if was_going_up and not going_up:
            charge_state = int(npoints / ((max_mz - min_mz) * (i - 1) + 0.5))
            current_score = autocorr_scores[i - 1]
            if abs(current_score / autocorr_scores[0]) > 0.05 and charge_state < max_charge:
                if abs(current_score) > best_autoscore:
                    best_autoscore = abs(current_score)
                    best_charge_state = charge_state
        was_going_up = going_up
    return best_autoscore, best_charge_state


@cython.nonecheck(False)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef tuple _autocorrelation_scores(object peak_index, object peak, double fwhm, double min_mz, double max_mz, size_t left_index, size_t right_index):
    """
    Compute an approximation of the fourier transform of the signals around the `peak`. The approximation
    of the fit is performed using the autocorrelation, which is then followed up with cubic spline.

    Parameters
    ----------
    peak_index : PeakIndex
        The signal container
    peak : FittedPeak
        The peak to fit around
    fwhm : float
        The area around the peak to fit
    min_mz : float
        The minimum m/z to search from
    max_mz : float
        The maximum m/z to search to
    left_index : int
        The index of the minimum m/z
    right_index : int
        The index of the maximum m/z

    Raises
    ------
    ValueError
        If not enough points can be found for spline interpolation

    Returns
    -------
    autoscores : np.ndarray(float64)
        The autocorrelation of the signals around `peak`
    num_l: int
        The number of points fitted over

    """

    cdef:
        object filtered
        double sum_of_differences, point_count, y1, y2, x1, x2, num_l_interp, desired_number, num_points
        size_t i
        long num_l
        np.ndarray[DTYPE_t, ndim=1] x, y, autoscores, intensity_array, mz_array
        DTYPE_t xval, yval

    filtered = peak_index
    sum_of_differences = 0.
    point_count = 0.

    mz_array = filtered.mz_array
    intensity_array = filtered.intensity_array

    for i in range(len(mz_array) - 1):
        y1 = intensity_array[i]
        y2 = intensity_array[i + 1]
        if y1 > 0 and y2 > 0:
            x1 = mz_array[i]
            x2 = mz_array[i + 1]

            sum_of_differences += x2 - x1
            point_count += 1
    if point_count > 5:
        average_differences = sum_of_differences / point_count

        num_l_interp = (math.ceil((max_mz - min_mz) / average_differences))
        num_l = int(num_l_interp + num_l_interp * 1.1)
    else:
        num_points = right_index - left_index + 1
        desired_number = 256
        point_multiplier = math.ceil(desired_number / num_points)

        if num_points < 5:
            raise ValueError("Not enough points")

        if num_points < desired_number:
            point_multiplier = max(5, point_multiplier)
            num_l = <long>(point_multiplier * num_points)
        else:
            num_l = <long>num_points

    interpolant = interpolate.interp1d(mz_array, intensity_array, kind='cubic',
                                       copy=False, assume_sorted=True, bounds_error=False)

    x = np.zeros(num_l)  # [0 for i in range(num_l)]
    y = np.zeros(num_l)

    for i in range(num_l):
        xval = (min_mz + ((max_mz - min_mz) * i) / num_l)
        yval = interpolant(xval)
        x[i] = xval
        y[i] = yval

    autoscores = autocorr(y)
    return autoscores, num_l



import operator
import numpy as np

from .utils import Base, ppm_error, range


class FittedPeak(Base):
    """Represent a single centroided mass spectral peak.

    FittedPeak instances are comparable for equality and
    hashable.

    Attributes
    ----------
    mz : float
        The m/z value at which the peak achieves its maximal
        abundance
    intensity : float
        The apex height of the peak
    signal_to_noise : float
        The signal to noise ratio of the peak
    full_width_at_half_max : double
        The symmetric full width at half of the
        peak's maximum height
    index : int
        The index at which the peak was found in the m/z array
        it was picked from
    peak_count : int
        The order in which the peak was picked
    area : float
        The total area of the peak, as determined
        by trapezoid integration
    left_width : float
        The left-sided width at half of max
    right_width : float
        The right-sided width at half of max
    """
    __slots__ = [
        "mz", "intensity", "signal_to_noise", "peak_count",
        "index", "full_width_at_half_max", "area",
        "left_width", "right_width"
    ]

    def __init__(self, mz, intensity, signal_to_noise, peak_count, index, full_width_at_half_max, area,
                 left_width=0, right_width=0):
        self.mz = mz
        self.intensity = intensity
        self.signal_to_noise = signal_to_noise
        self.peak_count = peak_count
        self.index = index
        self.full_width_at_half_max = full_width_at_half_max
        self.area = area
        self.left_width = left_width
        self.right_width = right_width

    def clone(self):
        """Creates a deep copy of the peak

        Returns
        -------
        FittedPeak
        """
        return FittedPeak(self.mz, self.intensity, self.signal_to_noise,
                          self.peak_count, self.index, self.full_width_at_half_max,
                          self.area, self.left_width, self.right_width)

    def __reduce__(self):
        return self.__class__, (self.mz, self.intensity, self.signal_to_noise,
                                self.peak_count, self.index, self.full_width_at_half_max,
                                self.area, self.left_width, self.right_width)

    def __hash__(self):
        return hash(self.mz)

    def __eq__(self, other):
        if other is None:
            return False
        return (abs(self.mz - other.mz) < 1e-5) and (
            abs(self.intensity - other.intensity) < 1e-5) and (
            abs(self.signal_to_noise - other.signal_to_noise) < 1e-5) and (
            abs(self.full_width_at_half_max - other.full_width_at_half_max) < 1e-5)

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return ("FittedPeak(mz=%0.3f, intensity=%0.3f, signal_to_noise=%0.3f, peak_count=%d, "
                "index=%d, full_width_at_half_max=%0.3f, area=%0.3f)") % (
                    self.mz, self.intensity, self.signal_to_noise,
                    self.peak_count, self.index, self.full_width_at_half_max,
                    self.area)


def _get_nearest_peak(peaklist, mz):
    lo = 0
    hi = len(peaklist)

    tol = 1

    def sweep(lo, hi):
        best_error = float('inf')
        best_index = None
        for i in range(hi - lo):
            i += lo
            v = peaklist[i].mz
            err = abs(v - mz)
            if err < best_error:
                best_error = err
                best_index = i
        return peaklist[best_index], best_error

    def binsearch(lo, hi):
        if (hi - lo) < 5:
            return sweep(lo, hi)
        else:
            mid = (hi + lo) // 2
            v = peaklist[mid].mz
            if abs(v - mz) < tol:
                return sweep(lo, hi)
            elif v > mz:
                return binsearch(lo, mid)
            else:
                return binsearch(mid, hi)
    return binsearch(lo, hi)


class PeakSet(Base):
    """A sequence of :class:`FittedPeak` instances, ordered by m/z,
    providing efficient search and retrieval of individual peaks or
    whole intervals of the m/z domain.

    This collection is not meant to be updated once created, as it
    it is indexed for ease of connecting individual peak objects to
    their position in the underlying sequence.

    One :class:`PeakSet` is considered equal to another if all of their
    contained :class:`FittedPeak` members are equal to each other

    Attributes
    ----------
    peaks : tuple
        The :class:`FittedPeak` instances, stored
    """

    def __init__(self, peaks):
        self.peaks = tuple(peaks)

    def __len__(self):
        return len(self.peaks)

    def reindex(self):
        """Re-indexes the sequence of peaks, updating their
        :attr:`peak_count` and setting their :attr:`index` if
        it is missing.
        """
        self._index()

    def _index(self):
        self.peaks = sorted(self.peaks, key=operator.attrgetter('mz'))
        i = 0
        for i, peak in enumerate(self.peaks):
            peak.peak_count = i
            if peak.index == -1:
                peak.index = i
        return i

    def has_peak(self, mz, tolerance=1e-5):
        """Search the for the peak nearest to `mz` within
        `tolerance` error (in PPM)

        Parameters
        ----------
        mz : float
            The m/z to search for
        tolerance : float, optional
            The error tolerance to accept. Defaults to 1e-5 (10 ppm)

        Returns
        -------
        FittedPeak
            The peak, if found. `None` otherwise.
        """
        return binary_search(self.peaks, mz, tolerance)

    get_nearest_peak = _get_nearest_peak

    def __repr__(self):
        return "<PeakSet %d Peaks>" % (len(self))

    def __getitem__(self, item):
        if isinstance(item, slice):
            return PeakSet(self.peaks[item])
        return self.peaks[item]

    def clone(self):
        """Creates a deep copy of the sequence of peaks

        Returns
        -------
        PeakSet
        """
        return PeakSet(p.clone() for p in self)

    def between(self, m1, m2):
        """Retrieve a :class:`PeakSet` containing all the peaks
        in `self` whose m/z is between `m1` and `m2`.

        These peaks are not copied.

        Parameters
        ----------
        m1 : float
            The lower m/z limit
        m2 : float
            The upper m/z limit

        Returns
        -------
        PeakSet
        """
        p1, _ = self.get_nearest_peak(m1)
        p2, _ = self.get_nearest_peak(m2)

        start = p1.peak_count
        end = p2.peak_count + 1
        n = len(self)
        if p1.mz < m1 and start + 1 < n:
            start += 1
        if p2.mz > m2 and end > 0:
            end -= 1
        if p2.mz < m1 or p1.mz > m2:
            end = start = 0
        return self[start:end]

    def all_peaks_for(self, mz, error_tolerance=2e-5):
        """Find all peaks within `error_tolerance` ppm of `mz` m/z.

        Parameters
        ----------
        mz : float
            The query m/z
        error_tolerance : float, optional
            The parts-per-million error tolerance (the default is 2e-5)

        Returns
        -------
        tuple
        """
        m1 = mz - (mz * error_tolerance)
        m2 = mz + (mz * error_tolerance)

        p1, _ = self.get_nearest_peak(m1)
        p2, _ = self.get_nearest_peak(m2)

        start = p1.peak_count
        end = p2.peak_count + 1
        n = len(self)
        if p1.mz < m1 and start + 1 < n:
            start += 1
        if p2.mz > m2 and end > 0:
            end -= 1
        if p2.mz < m1 or p1.mz > m2:
            end = start = 0
        return self.peaks[start:end]

    def __eq__(self, other):
        if other is None:
            return False
        return tuple(self) == tuple(other)

    def __ne__(self, other):
        return not (self == other)


def _sweep_solution(array, value, lo, hi, tolerance, verbose=False):
    best_index = -1
    best_error = float('inf')
    best_intensity = 0
    for i in range(hi - lo):
        target = array[lo + i]
        error = ppm_error(value, target.mz)
        abs_error = abs(error)
        if abs_error < tolerance and (abs_error < best_error * 1.1) and (target.intensity > best_intensity):
            best_index = lo + i
            best_error = abs_error
            best_intensity = target.intensity
    if best_index == -1:
        return None
    else:
        return array[best_index]


def _binary_search(array, value, lo, hi, tolerance, verbose=False):
    if (hi - lo) < 5:
        return _sweep_solution(array, value, lo, hi, tolerance, verbose)
    else:
        mid = (hi + lo) // 2
        target = array[mid]
        target_value = target.mz
        error = ppm_error(value, target_value)

        if abs(error) <= tolerance:
            return _sweep_solution(array, value, max(mid - (mid if mid < 5 else 5), lo), min(
                mid + 5, hi), tolerance, verbose)
        elif target_value > value:
            return _binary_search(array, value, lo, mid, tolerance, verbose)
        elif target_value < value:
            return _binary_search(array, value, mid, hi, tolerance, verbose)
    raise Exception("No recursion found!")


def binary_search(array, value, tolerance=2e-5, verbose=False):
    size = len(array)
    if size == 0:
        return None
    return _binary_search(array, value, 0, size, tolerance, verbose)


try:
    _has_c = True
    _FittedPeak = FittedPeak
    _PeakSet = PeakSet
    _p_binary_search = binary_search
    from ._c.peak_set import FittedPeak, PeakSetIndexed as PeakSet
except ImportError:
    _has_c = False


def to_array(peak_set):
    array = np.zeros((len(peak_set), 6))
    array[:, 0] = [p.mz for p in peak_set]
    array[:, 1] = [p.intensity for p in peak_set]
    array[:, 2] = [p.signal_to_noise for p in peak_set]
    array[:, 3] = [p.full_width_at_half_max for p in peak_set]
    array[:, 4] = [p.index for p in peak_set]
    array[:, 5] = [p.peak_count for p in peak_set]
    return array


def simple_peak(mz, intensity, fwhm=0.01):
    return FittedPeak(mz, intensity, intensity, -1, -1,
                      fwhm, intensity, fwhm / 2, fwhm / 2)


def is_peak(obj):
    return hasattr(obj, 'mz') and hasattr(obj, 'intensity')
